from leapp.actors import Actor
from leapp.models import DepCheck1, DepCheck2
from leapp.tags import SecondPhaseTag, WorkflowApiTestWorkflowTag
from leapp.workflows.api.v3 import DepCheckAPI4


class SecondPhaseActor(Actor):
    """
    This is just an actor to ensure no inpact on actors in different phases and that things like non tuple produces
    or consumes are handled properly.
    """

    name = 'second_phase_actor'
    consumes = (DepCheck1)  # On purpose not a tuple
    produces = (DepCheck2)  # On purpose not a tuple
    apis = (DepCheckAPI4)  # On purpose not a tuple
    tags = (SecondPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        pass



import itertools


def with_metaclass(meta_class, base_class=object):
    """
    :param meta_class: The desired metaclass to use
    :param base_class: The desired base class to use, the default one is object
    :type base_class: Type
    :return: Metaclass type to inherit from

    :Example:

    .. code-block:: python

       class MyMetaClass(type):
           def __new__(mcs, name, bases, attrs):
               klass = super(MyMetaClass, mcs).__new__(mcs, name, bases, attrs)
               klass.added = "Added field"
               return klass

       class MyClass(with_metaclass(MyMetaClass)):
           pass

       # This is equivalent to python 2:
       class MyClass(object):
           __metaclass__ = MyMetaClass

       # Or python 3
       class MyClass(object, metaclass=MyMetaClass):
           pass
    """
    return meta_class(
        'with_meta_base_' + base_class.__name__ + '_' + meta_class.__name__,
        (base_class,),
        {}
    )


def get_flattened_subclasses(cls):
    """
    Returns all the given subclasses and their subclasses recursively for the given class
    :param cls: Class to check
    :type cls: Type
    :return: Flattened list of subclasses and their subclasses
    """
    classes = cls.__subclasses__()
    return list(itertools.chain(classes, *[get_flattened_subclasses(x) for x in classes]))



from leapp.workflows.api.v3.testapi import TestAPI
from leapp.workflows.api.v2.newapi import NewAPI
from leapp.workflows.api.v1.depcheck import DepCheckAPI1, DepCheckAPI2, DepCheckAPI3, DepCheckAPI4




""" Leapp Models

These packages provide an interface describing a message payload data structure in the form of Models.
Together with the :py:mod:`leapp.models.fields`, package models are defined.

Example::

    class Boom(Model):
        topic = SomeTopic
        reason = fields.String()

    class Foobar(Model):
        topic = SomeTopic
        baz = fields.List(fields.String(), default=[])
        boom = fields.Model(Boom)

Now, the models can be used like this::

    f = Foobar(boom=Boom())
    f.boom.reason = "Example"
    f.baz.append("Add a string value to the list")
    from pprint import pprint
    pprint(f.dump())

"""
import sys

from leapp.models import fields
from leapp.models.error_severity import ErrorSeverity

from leapp.exceptions import ModelDefinitionError
from leapp.models.fields import ModelMisuseError
from leapp.utils.meta import get_flattened_subclasses, with_metaclass
from leapp.topics import ErrorTopic, DialogTopic, Topic


class ModelMeta(type):
    """
    ModelMeta is a metaclass used for Model

    It verifies the validity of attributes and registers the model as a message type with the :py:class:`Topic`.
    """
    def __new__(mcs, name, bases, attrs):
        klass = super(ModelMeta, mcs).__new__(mcs, name, bases, attrs)
        for base in bases:
            if getattr(base, '__non_inheritable__', False):
                raise TypeError('{cls} cannot inherit from {base} because {base} is not inheritable'.format(
                    cls=name, base=base.__name__))
        model_ref_cls = globals().get('_ModelReference')
        if name == '_ModelReference' or (model_ref_cls and issubclass(klass, model_ref_cls)):
            setattr(sys.modules[mcs.__module__], name, klass)
            return klass

        # Every model has to be bound to a topic except for the Model base class
        # Since it is using this class as meta class, we will dynamically get the class object from the globals
        model_cls = globals().get('Model')
        if model_cls and issubclass(klass, model_cls):
            topic = getattr(klass, 'topic', None)
            if not topic or not issubclass(topic, Topic):
                raise ModelDefinitionError('Missing topic in Model {}'.format(name))
            topic.messages = tuple(set(topic.messages + (klass,)))

        kls_attrs = (getattr(klass, 'fields', None) or {}).copy()
        kls_attrs.update({name: value for name, value in attrs.items() if isinstance(value, fields.Field)})
        klass.fields = kls_attrs.copy()

        setattr(sys.modules[mcs.__module__], name, klass)
        return klass

    def __init__(cls, name, bases, attrs):
        super(ModelMeta, cls).__init__(name, bases, attrs)


class Model(with_metaclass(ModelMeta)):
    """
    Model is a base class for all models.

    Models are defining the data structure of the payload of messages and the
    metadata required, such as a name and topic.
    """
    def __init__(self, init_method='from_initialization', **kwargs):
        super(Model, self).__init__()
        defined_fields = type(self).fields or {}
        for key in kwargs:
            if key not in defined_fields:
                raise ModelMisuseError(
                    'Trying to initialize model {} with value for undefined field {}'.format(type(self).__name__, key))
        for field in defined_fields:
            getattr(defined_fields[field], init_method)(kwargs, field, self)

    topic = None
    """
    `topic` has to be set to a subclass of :py:class:`leapp.topics.Topic`
    It defines the categorization of this model.
    """

    fields = None
    """
    `fields` contains a dictionary with all attributes of the py:class:`leapp.models.fields.Field` type in the class.

    Note: Dynamically added fields are ignored by the framework.
    """

    @classmethod
    def create(cls, data):
        """
        Create an instance of this class and use the data to initialize the fields within.

        :param data: Data to initialize the Model from deserialized data
        :type data: dict
        :return: Instance of this class
        """
        return cls(init_method='to_model', **data)

    def dump(self):
        """
        Dumps the data in the dictionary form that is safe to serialize to JSON.

        :return: dict with a builtin representation of the data that can be safely serialized to JSON
        """
        result = {}
        for field in type(self).fields.keys():
            type(self).fields[field].to_builtin(self, field, result)
        return result

    def __eq__(self, other):
        """
        Implementation for equality comparison of Model instances
        """
        return isinstance(other, type(self)) and \
            all(getattr(self, name) == getattr(other, name) for name in sorted(type(self).fields.keys()))

    @classmethod
    def serialize(cls):
        """ Returns serialized data of the model """
        return {
            'class_name': cls.__name__,
            'fields': {name: field.serialize() for name, field in cls.fields.items()},
            'topic': cls.topic.__name__
        }


class ErrorModel(Model):
    topic = ErrorTopic

    message = fields.String()
    severity = fields.StringEnum(choices=ErrorSeverity.ALLOWED_VALUES, default=ErrorSeverity.ERROR)
    details = fields.Nullable(fields.String())
    actor = fields.String()
    time = fields.DateTime()


class DialogModel(Model):
    topic = DialogTopic

    answerfile_sections = fields.JSON()
    actor = fields.String()
    details = fields.Nullable(fields.String())
    key = fields.Nullable(fields.String())


def get_models():
    """
    Returns a list of all currently loaded subclasses of Model

    :return: List Model subclasses
    """
    return [model for model in get_flattened_subclasses(Model) if model is not issubclass(model, _ModelReference)]


class _ModelReference(Model):
    _referenced = None
    _resolved = None

    def __new__(cls, *args, **kwargs):
        return cls.resolve()(*args, **kwargs)  # noqa; pylint: disable=not-callable

    @classmethod
    def create(cls, data):
        return cls.resolve()(init_method='to_model', **data)  # noqa; pylint: disable=not-callable

    @classmethod
    def resolve(cls):
        if not cls._resolved:
            try:
                cls._resolved = globals()[cls._referenced]
            except KeyError:
                raise ModelDefinitionError('Undefined Model "{}"'.format(cls._referenced))
        return cls._resolved


def resolve_model_references():
    """
    Resolves all dynamically created model references. When importing a model that has not been loaded
    yet, a dynamic model reference is created. After loading all models, resolve_model_references
    must be called to ensure the consistency of the code.

    :return: None
    """
    for reference in get_flattened_subclasses(_ModelReference):
        reference.resolve()


def _module_ref(name):
    reference = type(name + "Reference", (_ModelReference,), {})
    reference._referenced = name
    return reference


def _patch_module_getitem():
    class ReferenceDict(object):
        def __init__(self, module):
            self.__dict__['_module'] = module

        def __setattr__(self, name, value):
            setattr(self._module, name, value)

        def __delattr__(self, name):
            delattr(self._module, name)

        def __getattr__(self, item):
            return getattr(self._module, item, None) or _module_ref(item)

    sys.modules[__name__] = ReferenceDict(sys.modules[__name__])


_patch_module_getitem()



from __future__ import print_function
import ast
import functools
import itertools
import os


def print_section(data, section, pivot):
    "Print a section of obtained data"
    type_data = data[section]
    print('{}'.format(section.capitalize()))
    for td in type_data:
        fp = format_file_path(pivot, td['file'])
        first_part = '  - {}({})'.format(td['name'], ', '.join(td['bases']))
        pad = '.' * (60 - len(first_part))
        print('{} {} {}'.format(first_part, pad, fp))
    print('')


def format_file_path(pivot, path):
    "Format path as relative to a pivot"
    if not pivot or pivot == '.':
        pivot = os.getcwd()
    return os.path.relpath(path, pivot)


def get_candidate_files(start='.'):
    "Find all .py files in a directory tree"
    for root, unused, files in os.walk(start):
        for f in files:
            if not f.endswith('py'):
                continue
            yield os.path.join(root, f)


def ast_parse_file(filename):
    "Parse a python file and return tuple (ast, filename)"
    with open(filename, mode='r') as fp:
        try:
            return ast.parse(fp.read(), filename), filename
        except (SyntaxError, TypeError, ValueError):
            # Depending on python 3 version either TypeError or ValueError will be thrown if null bytes are
            # encountered
            return None, filename


def get_base_classes(bases, via):
    "Get base classes of a type, only direct names are supported currently"
    bases_set = set()
    errors = []
    for base in bases:
        if isinstance(base, ast.Name):
            bases_set.add(base.id)
        else:
            errors.append('Unknown base: {} via {}'.format(base.__class__.__name__, via))
    return bases_set, errors


def inspect(tree_file, collected_types=None, type_infos=None):
    "Inspect and collect data from AST tree"
    tree, filename = tree_file
    collected_types = collected_types or {}
    type_infos = type_infos or {}
    if not tree:
        return ['Unable to parse: {}'.format(filename)]
    errors = []
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            base_classes, err = get_base_classes(node.bases, filename)
            errors += err

            if base_classes & collected_types['models']:
                collected_types['models'].add(node.name)
                type_infos['models'].append({
                    'name': node.name,
                    'bases': list(base_classes),
                    'file': filename
                })
            if base_classes & collected_types['actors']:
                collected_types['actors'].add(node.name)
                type_infos['actors'].append({
                    'name': node.name,
                    'bases': list(base_classes),
                    'file': filename
                })
            if base_classes & collected_types['tags']:
                collected_types['tags'].add(node.name)
                type_infos['tags'].append({
                    'name': node.name,
                    'bases': list(base_classes),
                    'file': filename
                })
    return errors


def safe_discover(pivot):
    # Here we collect all the types that inherit from Model/Actor/Tag types
    # so that we can use this dict as a lookup for deeper search to support
    # use cases like:
    #
    # class FirstOrderModel(Model):
    #     pass
    #
    # class SecondOrderModel(FirstOrderModel):
    #     pass
    #
    collected_types = {
        'models': set(['Model']),
        'actors': set(['Actor']),
        'tags': set(['Tag'])
    }
    type_infos = {
        'models': [],
        'actors': [],
        'tags': []
    }

    inspector = functools.partial(
        inspect,
        collected_types=collected_types,
        type_infos=type_infos
    )

    errors = filter(None, map(inspector, map(ast_parse_file, get_candidate_files(pivot))))
    flat_errors = list(itertools.chain(*errors))

    print_section(type_infos, 'actors', pivot)
    print_section(type_infos, 'models', pivot)
    print_section(type_infos, 'tags', pivot)
    if flat_errors:
        print('Errors:')
        print('\n'.join(flat_errors))



import os
import socket

from leapp.utils.clicmd import command, command_opt
from leapp.cli import upgrade
from leapp import VERSION


@command('')
def cli(args): # noqa; pylint: disable=unused-argument
    pass


def main():
    os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()
    for cmd in [upgrade.list_runs, upgrade.preupgrade, upgrade.upgrade, upgrade.answer]:
        cli.command.add_sub(cmd.command)
    cli.command.execute('leapp version {}'.format(VERSION))



import os.path
import subprocess

import py


def test_workflow_api():
    repository_path = py.path.local(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'workflow-api-tests'))
    with repository_path.as_cwd():
        subprocess.check_call(['snactor', 'workflow', 'run', '--debug', 'WorkflowAPITest'])



import functools
import os
import pytest

from leapp.libraries.stdlib.call import _call


_CALLBACKS = [{}, [], 'string', lambda v: None, lambda a, b, c: None, None]
_POSITIVE_INTEGERS = [[], {}, 'x', -21, 0, None, -2.1, False, True, 0.0]
_STDIN = [[], {}, 0.123, lambda: None]


def test_invalid_command():
    with pytest.raises(TypeError):
        _call('i should be a list or tuple really')


def test_stdin_string():
    ret = _call(('bash', '-c', 'read MSG; echo "<$MSG>"'), stdin='LOREM IPSUM')
    assert ret['stdout'] == '<LOREM IPSUM>\n'


def test_stdin_fd():
    r, w = os.pipe()
    # The string we write here should not exceed `/proc/sys/fs/pipe-max-size`
    # which represents the size of the kernel buffer backing the pipe
    os.write(w, b'LOREM IPSUM')
    os.close(w)
    ret = _call(('bash', '-c', 'read MSG; echo "<$MSG>"'), stdin=r)
    os.close(r)
    assert ret['stdout'] == '<LOREM IPSUM>\n'


def test_linebuffer_callback():
    buffered = []

    def callback(fd, data):  # noqa; pylint: disable=unused-argument
        buffered.append(data)
    _call(('bash', '-c', 'echo 1; echo 2; echo 3'), callback_linebuffered=callback)
    assert buffered == ['1', '2', '3']


@pytest.mark.parametrize('p', _STDIN)
def test_stdin_check(p):
    with pytest.raises(TypeError):
        _call(('true',), stdin=p)


def test_output_1():
    ret = _call(('false',))
    assert isinstance(ret['exit_code'], int)
    assert ret['exit_code'] == 1
    assert isinstance(ret['pid'], int)
    assert ret['pid'] > 0
    assert ret['stdout'] == ''
    assert ret['stderr'] == ''
    assert isinstance(ret['signal'], int)
    assert ret['signal'] == 0


def test_output_2():
    ret = _call(('bash', '-c', 'echo STDOUT; (exec >&2 ; echo STDERR); exit 42',))
    assert isinstance(ret['exit_code'], int)
    assert ret['exit_code'] == 42
    assert isinstance(ret['pid'], int)
    assert ret['pid'] > 0
    assert ret['stdout'] == 'STDOUT\n'
    assert ret['stderr'] == 'STDERR\n'
    assert isinstance(ret['signal'], int)
    assert ret['signal'] == 0


def test_env_injection():
    ret = _call(('bash', '-c', 'echo $TEST'), env={'TEST': 'SUCCESS'})
    assert isinstance(ret['exit_code'], int)
    assert ret['exit_code'] == 0
    assert ret['stdout'] == 'SUCCESS\n'


def test_callability_check():
    with pytest.raises(TypeError):
        _call(('true',), callback_raw='nope')


@pytest.mark.parametrize('p', _CALLBACKS)
def test_callability_check_buffered(p):
    with pytest.raises(TypeError):
        _call(('true',), callback_linebuffered=p)


@pytest.mark.parametrize('p', _POSITIVE_INTEGERS)
def test_polltime(p):
    with pytest.raises(ValueError):
        _call(('true',), poll_timeout=p)


@pytest.mark.parametrize('p', _POSITIVE_INTEGERS)
def test_buffer_size(p):
    with pytest.raises(ValueError):
        _call(('true',), read_buffer_size=p)


class ArrayTracer(object):
    def __init__(self):
        self.value = []

    def __call__(self, unused, value):
        self.value.append(value)


def test_utf8_panagrams():
    panagrams_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'data',
        'call_data',
        'panagrams'
    )

    with open(panagrams_path, 'rb') as f:
        p_raw = f.read()
        panagrams = p_raw.decode('utf-8')

    primes = [
        2, 3, 5, 7, 11, 13, 17, 19,
        23, 29, 31, 37, 41, 43, 47, 53, 59,
        61, 67, 71, 73, 79, 83, 89, 97,
        101, 103, 107, 109, 113
    ]

    for prime in primes:
        _lines, _raw = ArrayTracer(), ArrayTracer()
        r = _call(
            ('cat', panagrams_path,), read_buffer_size=prime,
            callback_linebuffered=_lines, callback_raw=_raw
        )
        assert r['stdout'] == panagrams
        assert panagrams.splitlines() == _lines.value
        assert p_raw == bytes(functools.reduce(lambda x, xs: x + xs, _raw.value, bytes()))



class WorkflowAPI(object):
    produces = ()
    consumes = ()
    apis = ()

    def __init__(self):
        pass

    @classmethod
    def serialize(cls):
        return {
            'name': cls.__name__,
            'module': cls.__module__,
            'consumes': [model.__name__ for model in cls.consumes],
            'produces': [model.__name__ for model in cls.produces],
            'apis': [api.__name__ for api in cls.apis]
        }



import logging
import os
import socket
import sys
import uuid

from leapp.dialogs import RawMessageDialog
from leapp.exceptions import CommandError, MultipleConfigActorsError, WorkflowConfigNotAvailable
from leapp.messaging.answerstore import AnswerStore
from leapp.messaging.inprocess import InProcessMessaging
from leapp.messaging.commands import SkipPhasesUntilCommand
from leapp.tags import ExperimentalTag
from leapp.utils import reboot_system
from leapp.utils.audit import checkpoint, get_errors
from leapp.utils.meta import with_metaclass, get_flattened_subclasses
from leapp.utils.output import display_status_current_phase, display_status_current_actor
from leapp.workflows.phases import Phase
from leapp.workflows.policies import Policies
from leapp.workflows.phaseactors import PhaseActors


def _phase_sorter_key(a):
    return a.get_index()


def _is_phase(attr):
    return isinstance(attr, type) and issubclass(attr, Phase)


def _get_phases_sorted(attrs):
    return tuple(sorted([attr for attr in attrs.values() if _is_phase(attr)], key=_phase_sorter_key))


def actor_names(actor=None):
    return (actor.name.lower(), actor.class_name.lower()) if actor else ()


def phase_names(phase=None):
    return (phase[0].__name__.lower(), phase[0].name.lower()) if phase else ()


class WorkflowMeta(type):
    """
    Meta class for the registration of workflows
    """
    def __new__(mcs, name, bases, attrs):
        klass = super(WorkflowMeta, mcs).__new__(mcs, name, bases, attrs)
        if not getattr(sys.modules[mcs.__module__], name, None):
            setattr(sys.modules[mcs.__module__], name, klass)
        klass.phases = _get_phases_sorted(attrs)
        return klass


class _ConfigPhase(Phase):
    name = 'configuration_phase'


class Workflow(with_metaclass(WorkflowMeta)):
    """
    Workflow is the base class for all :ref:`workflow <terminology:workflow>` definitions.
    """

    name = None
    """Name of the workflow"""

    short_name = None
    """ Short name of the workflow """

    tag = None
    """ Workflow Tag """

    description = ''
    """ Documentation for the workflow """

    configuration = None
    """ Model to be used as workflow configuration """

    @property
    def errors(self):
        """
        :return: All reported errors
        """
        return self._errors

    @property
    def failure(self):
        return self._errors or self._unhandled_exception or self._stop_after_phase_requested

    @property
    def answer_store(self):
        """
        : return: AnswerStore instance used for messaging
        """
        return self._answer_store

    def save_answers(self, answerfile_path, userchoices_path):
        """
        Generates an answer file for the dialogs of the workflow and saves it to `answerfile_path`.
        Updates a .userchoices file at `userchoices_path` with new answers encountered in answerfile.

        :param answerfile_path: The path where to store the answer file.
        :param userchoices_path: The path where to store the .userchoices file.
        :return: None
        """
        # answerfile is generated only for the dialogs actually encountered in the worflow
        self._answer_store.generate(self._dialogs, answerfile_path)
        # userchoices is updated with any new data retrieved from answerfile
        self._answer_store.update(userchoices_path, allow_missing=True)

    def _load_from_file(self, filepath):
        if os.path.isfile(filepath):
            # XXX FIXME load_and_translate doesn't help here as somehow dialog.component.value
            # in Dialog.request_answers is not respectfully updated (set to None so storage
            # values are not taken into consideration).
            # Patching in 2 places - load here and direct call to translate in request_answers
            self._answer_store.load(filepath)
        else:
            self.log.warning("Previous file %s not found", filepath)

    def load_answers(self, answerfile_path, userchoices_path):
        self._load_from_file(userchoices_path)
        self._load_from_file(answerfile_path)

    def __init__(self, logger=None, auto_reboot=False):
        """
        :param logger: Optional logger to be used instead of leapp.workflow
        :type logger: Instance of :py:class:`logging.Logger`
        """
        self.log = (logger or logging.getLogger('leapp')).getChild('workflow')
        self._errors = []
        self._all_consumed = set()
        self._all_produced = set()
        self._initial = set()
        self._phase_actors = []
        self._experimental_whitelist = set()
        self._auto_reboot = auto_reboot
        self._unhandled_exception = False
        self._answer_store = AnswerStore()
        self._dialogs = []
        self._stop_after_phase_requested = False

        if self.configuration:
            config_actors = [actor for actor in self.tag.actors if self.configuration in actor.produces]
            if config_actors:
                if len(config_actors) == 1:
                    self._phase_actors.append((
                        _ConfigPhase,
                        PhaseActors((), 'Before'),
                        PhaseActors(tuple(config_actors), 'Main'),
                        PhaseActors((), 'After')))
                else:
                    config_actor_names = [a.name for a in config_actors]
                    raise MultipleConfigActorsError(config_actor_names)
        self.description = self.description or type(self).__doc__

        for phase in self.phases:
            phase.filter.tags += (self.tag,)
            self._phase_actors.append((
                phase,
                # filters all actors with the give tags
                # phasetag .Before
                self._apply_phase(phase.filter.get_before(), 'Before'),
                # phasetag
                self._apply_phase(phase.filter.get(), 'Main'),
                # phasetag .After
                self._apply_phase(phase.filter.get_after(), 'After')))

    def _apply_phase(self, actors, stage):
        phase_actors = PhaseActors(actors, stage)
        self._initial.update(set(phase_actors.initial) - self._all_produced)
        self._all_consumed.update(phase_actors.consumes)
        self._all_produced.update(phase_actors.produces)
        return phase_actors

    @property
    def experimental_whitelist(self):
        """ Whitelist of actors that may be executed even that they are marked experimental """
        return self._experimental_whitelist

    def whitelist_experimental_actor(self, actor):
        """
        Adds an actor to the experimental whitelist and allows them to be executed.

        :param actor: Actor to be whitelisted
        :type actor: class derived from py:class:`leapp.actors.Actor`
        :return: None
        """
        if actor:
            self._experimental_whitelist.add(actor)

    @property
    def phase_actors(self):
        """ Return all actors for the phase """
        return self._phase_actors

    @property
    def initial(self):
        """ Initial messages required """
        return self._initial

    @property
    def consumes(self):
        """ All consumed messages """
        return self._all_consumed

    @property
    def produces(self):
        """ All produced messages """
        return self._all_produced

    @property
    def dialogs(self):
        """ All encountered dialogs """
        return self._dialogs

    @classmethod
    def serialize(cls):
        """
        :return: Serialized form of the workflow
        """
        return {
            'name': cls.name,
            'short_name': cls.short_name,
            'tag': cls.tag.__name__,
            'description': cls.description,
            'phases': [phase.serialize() for phase in cls.phases],
        }

    def is_valid_phase(self, phase=None):
        if phase:
            return phase in [name for phs in self._phase_actors for name in phase_names(phs)]

    def run(self, context=None, until_phase=None, until_actor=None, skip_phases_until=None, skip_dialogs=False):
        """
        Executes the workflow

        :param context: Custom execution ID to be used instead of a randomly generated UUIDv4
        :type context: str
        :param until_phase: Specify until including which phase the execution should run - phase.stage can be used to
                            control it even more granularly. `phase` is any phase name where `stage` refers to `main`,
                            `before` or `after`. If no stage is defined, `after` is assumed to be the default value.
                            The execution ends when this phase (and stage, if specified) has been executed.
        :type until_phase: str
        :param until_actor: The execution finishes when this actor has been executed.
        :type until_actor: str
        :param skip_phases_until: Skips all phases until including the phase specified, and then continues the
               execution.
        :type skip_phases_until: str or None
        :param skip_dialogs: Inform actors about the mode of dialogs processing. If skip_dialogs is set to True it
                             means that dialogs can't be processed in the current workflow run interactively and
                             every attempted call of get_answers api method will be non-blocking, returning an empty
                             dict if no user choice was found in answerfile or a selected option otherwise.
                             If skip_dialogs is set to False then in case of absent recorded answer the dialog will
                             be rendered in a blocking user input requiring way.
                             The value of skip_dialogs will be passed to the actors that can theoretically use it for
                             their purposes.
        :type skip_dialogs: bool

        """
        context = context or str(uuid.uuid4())
        os.environ['LEAPP_EXECUTION_ID'] = context
        if not os.environ.get('LEAPP_HOSTNAME', None):
            os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()

        self.log.info('Starting workflow execution: {name} - ID: {id}'.format(
            name=self.name, id=os.environ['LEAPP_EXECUTION_ID']))

        skip_phases_until = (skip_phases_until or '').lower()
        needle_phase = until_phase or ''
        needle_stage = None
        if '.' in needle_phase:
            needle_phase, needle_stage = needle_phase.split('.', 1)
        needle_phase = needle_phase.lower()
        needle_stage = (needle_stage or '').lower()
        needle_actor = (until_actor or '').lower()

        self._errors = get_errors(context)
        config_model = type(self).configuration

        for phase in skip_phases_until, needle_phase:
            if phase and not self.is_valid_phase(phase):
                raise CommandError('Phase {phase} does not exist in the workflow'.format(phase=phase))

        self._stop_after_phase_requested = False
        for phase in self._phase_actors:
            os.environ['LEAPP_CURRENT_PHASE'] = phase[0].name
            if skip_phases_until:
                if skip_phases_until in phase_names(phase):
                    skip_phases_until = ''
                self.log.info('Skipping phase {name}'.format(name=phase[0].name))
                continue

            display_status_current_phase(phase)
            self.log.info('Starting phase {name}'.format(name=phase[0].name))
            current_logger = self.log.getChild(phase[0].name)

            early_finish = False
            for stage in phase[1:]:
                if early_finish:
                    return
                current_logger.info("Starting stage {stage} of phase {phase}".format(
                    phase=phase[0].name, stage=stage.stage))
                for actor in stage.actors:
                    if early_finish:
                        return
                    designation = ''
                    if ExperimentalTag in actor.tags:
                        designation = '[EXPERIMENTAL]'
                        if actor not in self.experimental_whitelist:
                            current_logger.info("Skipping experimental actor {actor}".format(actor=actor.name))
                            continue

                    display_status_current_actor(actor, designation=designation)
                    current_logger.info("Executing actor {actor} {designation}".format(designation=designation,
                                                                                       actor=actor.name))
                    messaging = InProcessMessaging(config_model=config_model, answer_store=self._answer_store)
                    messaging.load(actor.consumes)
                    instance = actor(logger=current_logger, messaging=messaging,
                                     config_model=config_model, skip_dialogs=skip_dialogs)
                    try:
                        instance.run()
                    except BaseException:
                        self._unhandled_exception = True
                        raise

                    self._stop_after_phase_requested = messaging.stop_after_phase or self._stop_after_phase_requested

                    # Collect dialogs
                    self._dialogs.extend(messaging.dialogs())
                    # Collect errors
                    if messaging.errors():
                        self._errors.extend(messaging.errors())

                        if phase[0].policies.error is Policies.Errors.FailImmediately:
                            self.log.info('Workflow interrupted due to FailImmediately error policy')
                            early_finish = True
                            break

                    for command in messaging.commands:
                        if command['command'] == SkipPhasesUntilCommand.COMMAND:
                            skip_phases_until = command['arguments']['until_phase']
                            self.log.info('SkipPhasesUntilCommand received. Skipping phases until {}'.format(
                                skip_phases_until))

                    checkpoint(actor=actor.name, phase=phase[0].name, context=context,
                               hostname=os.environ['LEAPP_HOSTNAME'])
                    if needle_actor in actor_names(actor):
                        self.log.info('Workflow finished due to the until-actor flag')
                        early_finish = True
                        break
                if not stage.actors:
                    checkpoint(actor='', phase=phase[0].name + '.' + stage.stage, context=context,
                               hostname=os.environ['LEAPP_HOSTNAME'])

                if needle_phase in phase_names(phase) and needle_stage == stage.stage.lower():
                    self.log.info('Workflow finished due to the until-phase flag')
                    early_finish = True
                    break

            checkpoint(actor='', phase=phase[0].name, context=context, hostname=os.environ['LEAPP_HOSTNAME'])

            if self._errors and phase[0].policies.error is Policies.Errors.FailPhase:
                self.log.info('Workflow interrupted due to the FailPhase error policy')
                early_finish = True

            elif needle_phase in phase_names(phase):
                self.log.info('Workflow finished due to the until-phase flag')
                early_finish = True

            elif self._stop_after_phase_requested:
                self.log.info('Workflow received request to stop after phase.')
                early_finish = True

            elif phase[0].flags.request_restart_after_phase or phase[0].flags.restart_after_phase:
                reboot = True
                if phase[0].flags.request_restart_after_phase and not self._auto_reboot:
                    reboot = False
                    messaging.request_answers(
                        RawMessageDialog(message='A reboot is required to continue. Please reboot your system.')
                    )
                if reboot:
                    self.log.info('Initiating system reboot due to the restart_after_reboot flag')
                    reboot_system()
                early_finish = True

            elif phase[0].flags.is_checkpoint:
                self.log.info('Stopping the workflow execution due to the is_checkpoint flag')
                early_finish = True

            if early_finish:
                return


def get_workflows():
    """
    :return: all registered workflows
    """
    return get_flattened_subclasses(Workflow)



#!/usr/bin/env python

import os

SCHEMA_DIR = os.path.dirname(__file__)
MIGRATIONS_DIR = os.path.join(SCHEMA_DIR, 'migrations')

MIGRATION_FILES = [filename for filename in os.listdir(MIGRATIONS_DIR) if
                   os.path.isfile(os.path.join(MIGRATIONS_DIR, filename))]


def migration_file_name_to_int(name):
    print('Filename {}'.format(name))
    name = name.split('-')[0]
    return int(name[:-1].lstrip('0') + name[-1])


def read_file(directory, path):
    with open(os.path.join(directory, path), 'r') as f:
        return f.read()


MIGRATION_DATA = [
    (migration_file_name_to_int(migration_file), read_file(MIGRATIONS_DIR, migration_file)) for
    migration_file in MIGRATION_FILES]
MIGRATION_DATA.sort()

SCHEMA_TEMPLATE = '''
""" This file is auto-generated by res/schema/embed.py - DO NOT MODIFY THIS FILE MANUALLY """

CURRENT_SCHEMA = """{current_schema}"""
"""Current schema of the leapp database"""

MIGRATIONS = [
    {migrations}
]
"""Migrations for previous versions of the leapp database"""
'''

with open(os.path.join(SCHEMA_DIR, 'schemas.py'), 'w') as f:
    f.write(SCHEMA_TEMPLATE.format(
        current_schema=read_file(SCHEMA_DIR, 'audit-layout.sql'),
        migrations=',\n    '.join('({}, """{}""")'.format(*entry) for entry in MIGRATION_DATA)))



import os
import json


def log_execution(actor):
    with open(os.environ['LEAPP_TEST_EXECUTION_LOG'], 'a+') as f:
        f.write(json.dumps(dict(name=actor.name, class_name=type(actor).__name__)) + '\n')



from __future__ import print_function
import itertools
import json
import os
import shutil
import sys
import tarfile
import uuid
from datetime import datetime

from leapp.config import get_config
from leapp.exceptions import CommandError, LeappError, UsageError
from leapp.logger import configure_logger
from leapp.messaging.answerstore import AnswerStore
from leapp.repository.scan import find_and_scan_repositories
from leapp.utils.audit import Execution, get_connection, get_checkpoints
from leapp.utils.clicmd import command, command_opt
from leapp.utils.output import (report_errors, report_info, beautify_actor_exception, report_unsupported,
                                report_inhibitors)
from leapp.utils.report import fetch_upgrade_report_messages, generate_report_file


def archive_logfiles():
    """ Archive log files from a previous run of Leapp """
    cfg = get_config()

    if not os.path.isdir(cfg.get('files_to_archive', 'dir')):
        os.makedirs(cfg.get('files_to_archive', 'dir'))

    files_to_archive = [os.path.join(cfg.get('files_to_archive', 'dir'), f)
                        for f in cfg.get('files_to_archive', 'files').split(',')
                        if os.path.isfile(os.path.join(cfg.get('files_to_archive', 'dir'), f))]

    if not os.path.isdir(cfg.get('archive', 'dir')):
        os.makedirs(cfg.get('archive', 'dir'))

    if files_to_archive:
        if os.path.isdir(cfg.get('debug', 'dir')):
            files_to_archive.append(cfg.get('debug', 'dir'))

        now = datetime.now().strftime('%Y%m%d%H%M%S')
        archive_file = os.path.join(cfg.get('archive', 'dir'), 'leapp-{}-logs.tar.gz'.format(now))

        with tarfile.open(archive_file, "w:gz") as tar:
            for file_to_add in files_to_archive:
                tar.add(file_to_add)
                if os.path.isdir(file_to_add):
                    shutil.rmtree(file_to_add, ignore_errors=True)
                try:
                    os.remove(file_to_add)
                except OSError:
                    pass
            # leapp_db is not in files_to_archive to not have it removed
            if os.path.isfile(cfg.get('database', 'path')):
                tar.add(cfg.get('database', 'path'))


def load_repositories_from(name, repo_path, manager=None):
    if get_config().has_option('repositories', name):
        repo_path = get_config().get('repositories', name)
    return find_and_scan_repositories(repo_path, manager=manager)


def load_repositories():
    manager = load_repositories_from('repo_path', '/etc/leapp/repo.d/', manager=None)
    manager.load()
    return manager


def fetch_last_upgrade_context():
    """
    :return: Context of the last execution
    """
    with get_connection(None) as db:
        cursor = db.execute(
            "SELECT context, stamp, configuration FROM execution WHERE kind = 'upgrade' ORDER BY id DESC LIMIT 1")
        row = cursor.fetchone()
        if row:
            return row[0], json.loads(row[2])
    return None, {}


def fetch_all_upgrade_contexts():
    """
    :return: All upgrade execution contexts
    """
    with get_connection(None) as db:
        cursor = db.execute(
            "SELECT context, stamp, configuration FROM execution WHERE kind = 'upgrade' ORDER BY id DESC")
        row = cursor.fetchall()
        if row:
            return row
    return None


def get_last_phase(context):
    checkpoints = get_checkpoints(context=context)
    if checkpoints:
        return checkpoints[-1]['phase']


def check_env_and_conf(env_var, conf_var, configuration):
    """
    Checks whether the given environment variable or the given configuration value are set to '1'
    """
    return os.getenv(env_var, '0') == '1' or configuration.get(conf_var, '0') == '1'


def generate_report_files(context):
    """
    Generates all report files for specific leapp run (txt and json format)
    """
    cfg = get_config()
    report_txt, report_json = [os.path.join(cfg.get('report', 'dir'),
                                            'leapp-report.{}'.format(f)) for f in ['txt', 'json']]
    # fetch all report messages as a list of dicts
    messages = fetch_upgrade_report_messages(context)
    generate_report_file(messages, context, report_json)
    generate_report_file(messages, context, report_txt)


def get_cfg_files(section, cfg, must_exist=True):
    """
    Provide files from particular config section
    """
    files = []
    for file_ in cfg.get(section, 'files').split(','):
        file_path = os.path.join(cfg.get(section, 'dir'), file_)
        if not must_exist or must_exist and os.path.isfile(file_path):
            files.append(file_path)
    return files


def warn_if_unsupported(configuration):
    env = os.environ
    if env.get('LEAPP_UNSUPPORTED', '0') == '1':
        devel_vars = {k: env[k] for k in env if k.startswith('LEAPP_DEVEL_')}
        report_unsupported(devel_vars, configuration["whitelist_experimental"])


def handle_output_level(args):
    """
    Set environment variables following command line arguments.
    """
    os.environ['LEAPP_DEBUG'] = '1' if args.debug else os.getenv('LEAPP_DEBUG', '0')
    if os.environ['LEAPP_DEBUG'] == '1' or args.verbose:
        os.environ['LEAPP_VERBOSE'] = '1'
    else:
        os.environ['LEAPP_VERBOSE'] = os.getenv('LEAPP_VERBOSE', '0')


def prepare_configuration(args):
    """Returns a configuration dict object while setting a few env vars as a side-effect"""
    if args.whitelist_experimental:
        args.whitelist_experimental = list(itertools.chain(*[i.split(',') for i in args.whitelist_experimental]))
        os.environ['LEAPP_EXPERIMENTAL'] = '1'
    else:
        os.environ['LEAPP_EXPERIMENTAL'] = '0'
    os.environ['LEAPP_UNSUPPORTED'] = '0' if os.getenv('LEAPP_UNSUPPORTED', '0') == '0' else '1'
    if args.no_rhsm:
        os.environ['LEAPP_NO_RHSM'] = '1'
    elif os.getenv('LEAPP_NO_RHSM') != '1':
        os.environ['LEAPP_NO_RHSM'] = os.getenv('LEAPP_DEVEL_SKIP_RHSM', '0')
    if args.enablerepo:
        os.environ['LEAPP_ENABLE_REPOS'] = ','.join(args.enablerepo)
    configuration = {
        'debug': os.getenv('LEAPP_DEBUG', '0'),
        'verbose': os.getenv('LEAPP_VERBOSE', '0'),
        'whitelist_experimental': args.whitelist_experimental or (),
    }
    return configuration


def process_whitelist_experimental(repositories, workflow, configuration, logger=None):
    for actor_name in configuration.get('whitelist_experimental', ()):
        actor = repositories.lookup_actor(actor_name)
        if actor:
            workflow.whitelist_experimental_actor(actor)
        else:
            msg = 'No such Actor: {}'.format(actor_name)
            if logger:
                logger.error(msg)
            raise CommandError(msg)


@command('upgrade', help='Upgrade the current system to the next available major version.')
@command_opt('resume', is_flag=True, help='Continue the last execution after it was stopped (e.g. after reboot)')
@command_opt('reboot', is_flag=True, help='Automatically performs reboot when requested.')
@command_opt('whitelist-experimental', action='append', metavar='ActorName', help='Enable experimental actors')
@command_opt('debug', is_flag=True, help='Enable debug mode', inherit=False)
@command_opt('verbose', is_flag=True, help='Enable verbose logging', inherit=False)
@command_opt('no-rhsm', is_flag=True, help='Use only custom repositories and skip actions'
                                           ' with Red Hat Subscription Manager')
@command_opt('enablerepo', action='append', metavar='<repoid>',
             help='Enable specified repository. Can be used multiple times.')
def upgrade(args):
    skip_phases_until = None
    context = str(uuid.uuid4())
    cfg = get_config()
    handle_output_level(args)
    configuration = prepare_configuration(args)
    answerfile_path = cfg.get('report', 'answerfile')
    userchoices_path = cfg.get('report', 'userchoices')

    if os.getuid():
        raise CommandError('This command has to be run under the root user.')

    if args.resume:
        context, configuration = fetch_last_upgrade_context()
        if not context:
            raise CommandError('No previous upgrade run to continue, remove `--resume` from leapp invocation to'
                               ' start a new upgrade flow')
        os.environ['LEAPP_DEBUG'] = '1' if check_env_and_conf('LEAPP_DEBUG', 'debug', configuration) else '0'

        if os.environ['LEAPP_DEBUG'] == '1' or check_env_and_conf('LEAPP_VERBOSE', 'verbose', configuration):
            os.environ['LEAPP_VERBOSE'] = '1'
        else:
            os.environ['LEAPP_VERBOSE'] = '0'

        skip_phases_until = get_last_phase(context)
        logger = configure_logger()
    else:
        e = Execution(context=context, kind='upgrade', configuration=configuration)
        e.store()
        archive_logfiles()
        logger = configure_logger('leapp-upgrade.log')
    os.environ['LEAPP_EXECUTION_ID'] = context

    if args.resume:
        logger.info("Resuming execution after phase: %s", skip_phases_until)
    try:
        repositories = load_repositories()
    except LeappError as exc:
        raise CommandError(exc.message)
    workflow = repositories.lookup_workflow('IPUWorkflow')(auto_reboot=args.reboot)
    process_whitelist_experimental(repositories, workflow, configuration, logger)
    warn_if_unsupported(configuration)
    with beautify_actor_exception():
        logger.info("Using answerfile at %s", answerfile_path)
        workflow.load_answers(answerfile_path, userchoices_path)
        workflow.run(context=context, skip_phases_until=skip_phases_until, skip_dialogs=True)

    logger.info("Answerfile will be created at %s", answerfile_path)
    workflow.save_answers(answerfile_path, userchoices_path)
    report_errors(workflow.errors)
    report_inhibitors(context)
    generate_report_files(context)
    report_files = get_cfg_files('report', cfg)
    log_files = get_cfg_files('logs', cfg)
    report_info(report_files, log_files, answerfile_path, fail=workflow.failure)

    if workflow.failure:
        sys.exit(1)


@command('preupgrade', help='Generate preupgrade report')
@command_opt('whitelist-experimental', action='append', metavar='ActorName', help='Enables experimental actors')
@command_opt('debug', is_flag=True, help='Enable debug mode', inherit=False)
@command_opt('verbose', is_flag=True, help='Enable verbose logging', inherit=False)
@command_opt('no-rhsm', is_flag=True, help='Use only custom repositories and skip actions'
                                           ' with Red Hat Subscription Manager')
@command_opt('enablerepo', action='append', metavar='<repoid>',
             help='Enable specified repository. Can be used multiple times.')
def preupgrade(args):
    context = str(uuid.uuid4())
    cfg = get_config()
    handle_output_level(args)
    configuration = prepare_configuration(args)
    answerfile_path = cfg.get('report', 'answerfile')
    userchoices_path = cfg.get('report', 'userchoices')

    if os.getuid():
        raise CommandError('This command has to be run under the root user.')
    e = Execution(context=context, kind='preupgrade', configuration=configuration)
    e.store()
    archive_logfiles()
    logger = configure_logger('leapp-preupgrade.log')
    os.environ['LEAPP_EXECUTION_ID'] = context

    try:
        repositories = load_repositories()
    except LeappError as exc:
        raise CommandError(exc.message)
    workflow = repositories.lookup_workflow('IPUWorkflow')()
    warn_if_unsupported(configuration)
    process_whitelist_experimental(repositories, workflow, configuration, logger)
    with beautify_actor_exception():
        workflow.load_answers(answerfile_path, userchoices_path)
        until_phase = 'ReportsPhase'
        logger.info('Executing workflow until phase: %s', until_phase)
        workflow.run(context=context, until_phase=until_phase, skip_dialogs=True)

    logger.info("Answerfile will be created at %s", answerfile_path)
    workflow.save_answers(answerfile_path, userchoices_path)
    generate_report_files(context)
    report_errors(workflow.errors)
    report_inhibitors(context)
    report_files = get_cfg_files('report', cfg)
    log_files = get_cfg_files('logs', cfg)
    report_info(report_files, log_files, answerfile_path, fail=workflow.failure)
    if workflow.failure:
        sys.exit(1)


@command('answer', help='Manage answerfile generation: register persistent user choices for specific dialog sections')
@command_opt('section', action='append', metavar='dialog_sections',
             help='Register answer for a specific section in the answerfile')
@command_opt('add', is_flag=True,
             help='If set sections will be created even if missing in original answerfile')
def answer(args):
    """A command to record user choices to the questions in the answerfile.
       Saves user answer between leapp preupgrade runs.
    """
    cfg = get_config()
    if args.section:
        args.section = list(itertools.chain(*[i.split(',') for i in args.section]))
    else:
        raise UsageError('At least one dialog section must be specified, ex. --section dialog.option=mychoice')
    try:
        sections = [tuple((dialog_option.split('.', 2) + [value]))
                    for dialog_option, value in [s.split('=', 2) for s in args.section]]
    except ValueError:
        raise UsageError("A bad formatted section has been passed. Expected format is dialog.option=mychoice")
    answerfile_path = cfg.get('report', 'answerfile')
    answerstore = AnswerStore()
    answerstore.load(answerfile_path)
    for dialog, option, value in sections:
        answerstore.answer(dialog, option, value)
    not_updated = answerstore.update(answerfile_path, allow_missing=args.add)
    if not_updated:
        sys.stderr.write("WARNING: Only sections found in original userfile can be updated, ignoring {}\n".format(
            ",".join(not_updated)))


@command('list-runs', help='List previous Leapp upgrade executions')
def list_runs(args):  # noqa; pylint: disable=unused-argument
    contexts = fetch_all_upgrade_contexts()
    if contexts:
        for context in contexts:
            print('Context ID: {} - time: {} - details: {}'.format(context[0], context[1], json.loads(context[2])),
                  file=sys.stdout)
    else:
        raise CommandError('No previous run found!')



from leapp.models import UnitTestConfig
from leapp.workflows import Workflow
from leapp.workflows.phases import Phase
from leapp.workflows.flags import Flags
from leapp.workflows.tagfilters import TagFilter
from leapp.workflows.policies import Policies
from leapp.tags import UnitTestWorkflowTag, FirstPhaseTag, SecondPhaseTag, ThirdPhaseTag, FourthPhaseTag, FifthPhaseTag


class UnitTestWorkflow(Workflow):
    name = 'UnitTest'
    tag =  UnitTestWorkflowTag
    short_name = 'unit_test'
    description = '''No description has been provided for the UnitTest workflow.'''
    configuration = UnitTestConfig

    class FirstPhase(Phase):
       name = 'first-phase'
       filter = TagFilter(FirstPhaseTag)
       policies = Policies(Policies.Errors.FailImmediately,
                           Policies.Retry.Phase)
       flags = Flags()

    class SecondPhase(Phase):
       name = 'second-phase'
       filter = TagFilter(SecondPhaseTag)
       policies = Policies(Policies.Errors.FailPhase,
                           Policies.Retry.Phase)
       flags = Flags()

    class ThirdPhase(Phase):
       name = 'third-phase'
       filter = TagFilter(ThirdPhaseTag)
       policies = Policies(Policies.Errors.FailPhase,
                           Policies.Retry.Phase)
       flags = Flags()

    class FourthPhase(Phase):
       name = 'fourth-phase'
       filter = TagFilter(FourthPhaseTag)
       policies = Policies(Policies.Errors.FailPhase,
                           Policies.Retry.Phase)
       flags = Flags()

    class FifthPhase(Phase):
       name = 'fifth-phase'
       filter = TagFilter(FifthPhaseTag)
       policies = Policies(Policies.Errors.FailPhase,
                           Policies.Retry.Phase)
       flags = Flags()

    # Template for phase definition - The order in which the phase classes are defined
    # within the Workflow class represents the execution order
    #
    # class PhaseName(Phase):
    #    name = 'phase_name'
    #    filter = TagFilter(PhaseTag)
    #    policies = Policies(Policies.Errors.FailPhase,
    #                        Policies.Retry.Phase)
    #    flags = Flags()



import logging

import mock
import pytest

from leapp.repository.actor_definition import ActorDefinition, ActorInspectionFailedError, MultipleActorsError
from leapp.exceptions import UnsupportedDefinitionKindError
from leapp.repository import DefinitionKind

_FAKE_META_DATA = {
    'description': 'Fake Description',
    'class_name': 'FakeActor',
    'name': 'fake-actor',
    'path': 'actors/test',
    'tags': (),
    'consumes': (),
    'produces': (),
    'dialogs': (),
    'apis': ()
}


def test_actor_definition(repository_dir):
    with repository_dir.as_cwd():
        logger = logging.getLogger('leapp.actor.test')
        with mock.patch.object(logger, 'log') as log_mock:
            definition = ActorDefinition('actors/test', '.', log=log_mock)
            for kind in set(DefinitionKind.REPO_WHITELIST + DefinitionKind.ACTOR_WHITELIST):
                if kind in DefinitionKind.ACTOR_WHITELIST:
                    definition.add(kind, '.')
                else:
                    with pytest.raises(UnsupportedDefinitionKindError):
                        definition.add(kind, '.')
                    log_mock.error.assert_called_with(
                        "Attempt to add item type %s to actor that is not supported", kind.name)
                    log_mock.reset_mock()
            with mock.patch('leapp.repository.actor_definition.get_actor_metadata', return_value=_FAKE_META_DATA):
                with mock.patch('leapp.repository.actor_definition.get_actors', return_value=[True]):
                    definition._module = True
                    assert definition.consumes == _FAKE_META_DATA['consumes']
                    assert definition.produces == _FAKE_META_DATA['produces']
                    assert definition.tags == _FAKE_META_DATA['tags']
                    assert definition.class_name == _FAKE_META_DATA['class_name']
                    assert definition.dialogs == _FAKE_META_DATA['dialogs']
                    assert definition.name == _FAKE_META_DATA['name']
                    assert definition.description == _FAKE_META_DATA['description']
                    assert definition.apis == _FAKE_META_DATA['apis']
                    dumped = definition.serialize()
                    assert dumped.pop('class_name') == definition.class_name
                    assert dumped.pop('description') == definition.description
                    assert dumped.pop('consumes') == definition.consumes
                    assert dumped.pop('produces') == definition.produces
                    assert dumped.pop('apis') == definition.apis
                    assert dumped.pop('tags') == definition.tags
                    assert dumped.pop('dialogs') == [dialog.serialize() for dialog in definition.dialogs]
                    assert dumped.pop('path') == _FAKE_META_DATA['path']
                    assert dumped.pop('name') == definition.name
                    assert dumped.pop('files') == ('.',)
                    assert dumped.pop('libraries') == ('.',)
                    assert dumped.pop('tests') == ('.',)
                    assert dumped.pop('tools') == ('.',)
                    # Assert to ensure we covered all keys
                    assert not dumped

            with pytest.raises(ActorInspectionFailedError):
                with mock.patch('leapp.repository.actor_definition.get_actors', return_value=[]):
                    definition._discovery = None
                    definition.discover()

            with pytest.raises(ActorInspectionFailedError):
                with mock.patch('leapp.repository.actor_definition.get_actors') as mocked_actors:
                    mocked_actors.side_effect = RuntimeError('Test error')
                    definition._discovery = None
                    definition.discover()

            with pytest.raises(MultipleActorsError):
                with mock.patch('leapp.repository.actor_definition.get_actor_metadata', return_value=_FAKE_META_DATA):
                    with mock.patch('leapp.repository.actor_definition.get_actors', return_value=[True, True]):
                        definition._discovery = None
                        definition.discover()



from leapp.actors import Actor
from leapp.tags import FirstPhaseTag, WorkflowApiTestWorkflowTag
from leapp.workflows.api.v2 import TestAPI


class ApiV2Test(Actor):
    """
    This actor will check that the testing API ApiV3Test is behaving as expected.
    """

    name = 'api_v2_test'
    consumes = ()
    produces = ()
    apis = (TestAPI,)
    tags = (FirstPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        api = TestAPI()
        # Expected output is always the order a, b, c
        # The parameter order here is however c, b, a
        if api.order_change(3, 2, 1) != (1, 2, 3):
            raise StopActorExecutionError("order change API failure")

        # Expected is that the input is prefixed with 'survivor.v3.'
        if api.survivor("APIV2Test") != "survivor.v3.APIV2Test":
            raise StopActorExecutionError("v2 survivor test failure")



from leapp.workflows.api import WorkflowAPI


class TestAPI(WorkflowAPI):
    """
    V3 implementation of the TestAPI
    """

    def order_change(self, b, c, a):
        """ 'Latest implementation' of the api """
        print("v3.TestAPI.order_change => " + str((a, b, c)))
        # Returns a, b, c as promised
        return a, b, c

    def survivor(self, value):
        """ The actual prefixing implementation """
        print("v3.TestAPI.survivor")
        # Prefixes the value and returns it
        return "survivor.v3." + value



from __future__ import print_function

import hashlib
import json
import os
import sys
from contextlib import contextmanager

from leapp.exceptions import LeappRuntimeError
from leapp.models import ErrorModel
from leapp.utils.audit import get_audit_entry


def _is_verbose():
    """Redefinition of is_verbose from leapp.libraries.stdlib.config because it leads to import errors"""
    return os.getenv('LEAPP_VERBOSE', '0') == '1'


class Color(object):
    reset = "\033[0m" if sys.stdout.isatty() else ""
    bold = "\033[1m" if sys.stdout.isatty() else ""
    red = "\033[1;31m" if sys.stdout.isatty() else ""
    green = "\033[1;32m" if sys.stdout.isatty() else ""
    yellow = "\033[1;33m" if sys.stdout.isatty() else ""


def pretty_block_text(string, color=Color.bold, width=60):
    return "\n{color}{separator}\n{text}\n{separator}{reset}\n".format(
        color=color,
        separator="=" * width,
        reset=Color.reset,
        text=string.center(width))


@contextmanager
def pretty_block(text, target, end_text=None, color=Color.bold, width=60):
    target.write(pretty_block_text(text, color=color, width=width))
    target.write('\n')
    yield
    target.write(pretty_block_text(end_text or 'END OF {}'.format(text), color=color, width=width))
    target.write('\n')


def print_error(error):
    model = ErrorModel.create(json.loads(error['message']['data']))
    sys.stdout.write("{red}{time} [{severity}]{reset} Actor: {actor}\nMessage: {message}\n".format(
        red=Color.red, reset=Color.reset, severity=model.severity.upper(),
        message=model.message, time=model.time, actor=model.actor))
    if model.details:
        print('Summary:')
        details = json.loads(model.details)
        for detail in details:
            print('    {k}: {v}'.format(
                k=detail.capitalize(),
                v=details[detail].rstrip().replace('\n', '\n' + ' ' * (6 + len(detail)))))


def report_inhibitors(context_id):
    # The following imports are required to be here to avoid import loop problems
    from leapp.reporting import Flags  # pylint: disable=import-outside-toplevel
    from leapp.utils.report import fetch_upgrade_report_messages  # pylint: disable=import-outside-toplevel
    reports = fetch_upgrade_report_messages(context_id)
    inhibitors = [report for report in reports if Flags.INHIBITOR in report.get('flags', [])]
    if inhibitors:
        text = 'UPGRADE INHIBITED'
        with pretty_block(text=text, end_text=text, color=Color.red, target=sys.stdout):
            print('Upgrade has been inhibited due to the following problems:')
            for position, report in enumerate(inhibitors, start=1):
                print('{idx:5}. Inhibitor: {title}'.format(idx=position, title=report['title']))
            print('Consult the pre-upgrade report for details and possible remediation.')


def report_deprecations(context_id, start=None):
    deprecations = get_audit_entry(event='deprecation', context=context_id)
    if start:
        start_stamp = start.isoformat() + 'Z'
        deprecations = [d for d in deprecations if d['stamp'] > start_stamp]
    if deprecations:
        cache = set()
        with pretty_block("USE OF DEPRECATED ENTITIES", target=sys.stderr, color=Color.red):
            for deprecation in deprecations:
                entry_data = json.loads(deprecation['data'])
                # Deduplicate messages
                key = hashlib.sha256(json.dumps(entry_data, sort_keys=True)).hexdigest()
                if key in cache:
                    continue
                # Add current message to the cache
                cache.add(key)
                # Print the message
                sys.stderr.write(
                    '{message} @ {filename}:{lineno}\nNear: {line}\nReason: {reason}\n{separator}\n'.format(
                        separator='-' * 60, **entry_data)
                )


def report_errors(errors):
    if errors:
        with pretty_block("ERRORS", target=sys.stdout, color=Color.red):
            for error in errors:
                print_error(error)


def report_info(report_paths, log_paths, answerfile=None, fail=False):
    report_paths = [report_paths] if not isinstance(report_paths, list) else report_paths
    log_paths = [log_paths] if not isinstance(report_paths, list) else log_paths

    if log_paths:
        sys.stdout.write("\n")
        for log_path in log_paths:
            sys.stdout.write("Debug output written to {path}\n".format(path=log_path))

    if report_paths:
        with pretty_block("REPORT", target=sys.stdout, color=Color.bold if fail else Color.green):
            for report_path in report_paths:
                sys.stdout.write("A report has been generated at {path}\n".format(path=report_path))

    if answerfile:
        sys.stdout.write("Answerfile has been generated at {}\n".format(answerfile))


def report_unsupported(devel_vars, experimental):
    text = "UNSUPPORTED UPGRADE"
    with pretty_block(text=text, end_text=text, target=sys.stdout, color=Color.yellow):
        sys.stdout.write("Variable LEAPP_UNSUPPORTED has been detected. Proceeding at your own risk.\n")

        if devel_vars:
            sys.stdout.write("{yellow}Development variables{reset} have been detected:\n".format(
                yellow=Color.yellow, reset=Color.reset))
            for key in devel_vars:
                sys.stdout.write("- {key}={value}\n".format(key=key, value=devel_vars[key]))
        if experimental:
            sys.stdout.write("{yellow}Experimental actors{reset} have been detected:\n".format(
                yellow=Color.yellow, reset=Color.reset))
            for actor in experimental:
                sys.stdout.write("- {actor}\n".format(actor=actor))


@contextmanager
def beautify_actor_exception():
    try:
        try:
            yield
        except LeappRuntimeError as e:
            msg = '{} - Please check the above details'.format(e.message)
            sys.stderr.write("\n")
            sys.stderr.write(pretty_block_text(msg, color="", width=len(msg)))
    finally:
        pass


def display_status_current_phase(phase):
    if not _is_verbose():
        print('==> Processing phase `{name}`'.format(name=phase[0].name))


def _get_description_title(actor):
    lines = actor.description.strip().split('\n')
    return lines[0].strip() if lines else actor.name


def display_status_current_actor(actor, designation=''):
    if not _is_verbose():
        print('====> * {actor}{designation}\n        {title}'.format(actor=actor.name,
                                                                     title=_get_description_title(actor),
                                                                     designation=designation))



import os
import pkgutil
import socket

from leapp.utils.i18n import _  # noqa; pylint: disable=redefined-builtin
from leapp.snactor import commands
from leapp.snactor.commands import workflow
from leapp.snactor.commands import messages
from leapp.snactor.commands import repo
from leapp.utils.clicmd import command, command_opt
from leapp.utils.repository import find_repository_basedir
from leapp import VERSION

SHORT_HELP = _("snactor is a development and repository management tool for Leapp.")
LONG_HELP = _("""Snactor is designed to get quickly started with leapp actor development.""")


def load_commands():
    _load_commands_from(commands.__file__)
    _load_commands_from(commands.workflow.__file__)
    _load_commands_from(commands.messages.__file__)
    _load_commands_from(commands.repo.__file__)
    cli.command.add_sub(messages.messages.command)
    cli.command.add_sub(workflow.workflow.command)
    cli.command.add_sub(repo.repo.command)


def _load_commands_from(path):
    pkg_path = os.path.dirname(path)
    for importer, name, is_pkg in pkgutil.iter_modules([pkg_path]):
        if is_pkg:
            continue
        mod = importer.find_module(name).load_module(name)
        if hasattr(mod.cli, 'command'):
            if not mod.cli.command.parent:
                cli.command.add_sub(mod.cli.command)


@command('', help=LONG_HELP)
@command_opt('debug', is_flag=True, help=_('Enables debug mode'), inherit=True)
@command_opt('verbose', is_flag=True, help=_('Enables verbose logging'), inherit=True)
@command_opt('config', help=_('Allows to override the leapp.conf location'), inherit=True)
@command_opt('logger-config', help=_('Allows to override the logger.conf location'), inherit=True)
def cli(args):
    if args.logger_config and os.path.isfile(args.logger_config):
        os.environ['LEAPP_LOGGER_CONFIG'] = args.logger_config
    # Consider using the in repository $REPOPATH/.leapp/logger.conf to actually obey --debug / --verbose
    # If /etc/leapp/logger.conf or $REPOPATH/.leapp/logger.conf don't exist logging won't work in snactor.
    elif find_repository_basedir('.') and os.path.isfile(os.path.join(find_repository_basedir('.'),
                                                                      '.leapp/logger.conf')):
        os.environ['LEAPP_LOGGER_CONFIG'] = os.path.join(find_repository_basedir('.'), '.leapp/logger.conf')

    config_file_path = None
    if args.config and os.path.isfile(args.config):
        config_file_path = args.config

    if not config_file_path and find_repository_basedir('.'):
        config_file_path = os.path.join(find_repository_basedir('.'), '.leapp/leapp.conf')

    if not config_file_path or not os.path.isfile(config_file_path):
        config_file_path = os.environ.get('LEAPP_CONFIG')

    if not config_file_path or not os.path.isfile(config_file_path):
        config_file_path = '/etc/leapp/leapp.conf'

    os.environ['LEAPP_CONFIG'] = config_file_path
    os.environ['LEAPP_DEBUG'] = '1' if args.debug else os.environ.get('LEAPP_DEBUG', '0')

    if os.environ['LEAPP_DEBUG'] == '1' or args.verbose:
        os.environ['LEAPP_VERBOSE'] = '1'
    else:
        os.environ['LEAPP_VERBOSE'] = os.environ.get('LEAPP_VERBOSE', '0')


def main():
    os.environ['LEAPP_HOSTNAME'] = socket.getfqdn()
    load_commands()
    cli.command.execute(version=_('snactor version {}').format(VERSION))



import gettext  # noqa: F401; pylint: disable=unused-import
import locale
import sys


__all__ = ('string_types', 'IS_PYTHON2', 'IS_PYTHON3', 'httplib', 'unicode_type', 'raise_with_traceback')

IS_PYTHON2 = sys.version_info < (3,)
IS_PYTHON3 = not IS_PYTHON2

string_types = ()


# Python 2 code
if IS_PYTHON2:

    import httplib
    string_types = (str, globals()['__builtins__']['unicode'])
    unicode_type = string_types[1]
    from leapp.compatpy2only import raise_with_traceback
    builtins_dict = globals()['__builtins__']

    def gettext_setup(t):
        def us(u):
            if isinstance(u, globals()['__builtins__']['unicode']):
                return u
            return unicode_type(u)

        def singular(msg):
            return t.ugettext(us(msg))

        def plural(msg1, msg2, n):
            return t.ungettext(us(msg1), us(msg2), n)

        return singular, plural

    def setlocale(category, loc=None):
        locale.setlocale(category, loc.encode('utf-8') if loc else None)

# Python 3 code
else:
    import http.client as httplib
    import builtins
    string_types = (str,)
    unicode_type = str
    builtins_dict = builtins.__dict__

    def gettext_setup(t):
        singular = t.gettext
        plural = t.ngettext
        return singular, plural

    def setlocale(category, loc=None):
        locale.setlocale(category, loc)

    def raise_with_traceback(exc, tb):
        """
        This is a helper function to raise exceptions with a traceback.

        This is function is required to workaround the syntax changes between Python 2 and 3
        Python 3.4 introduced a with_traceback method to Exception classes and Python 3 removed the syntax
        which used to be used in Python 2.

        :param exc: Exception to raise
        :param tb:  Traceback to use
        :return: Nothing
        """
        raise exc.with_traceback(tb)



from __future__ import print_function

import datetime
import os
import sys

from leapp.exceptions import CommandError, LeappError
from leapp.logger import configure_logger
from leapp.repository.scan import find_and_scan_repositories
from leapp.snactor.commands.workflow import workflow
from leapp.snactor.context import with_snactor_context
from leapp.utils.clicmd import command_arg, command_opt
from leapp.utils.output import beautify_actor_exception, report_deprecations, report_errors
from leapp.utils.repository import find_repository_basedir, requires_repository

_LONG_DESCRIPTION = '''
Executes the given workflow.

Using --until-phase the workflow will be only executed until including
the given phase.

Using --until-actor the workflow will be only executed until including
the first occurrence of the given actor name.

For more information please consider reading the documentation at:
https://red.ht/leapp-docs
'''


@workflow.command('run', help='Execute a workflow with the given name', description=_LONG_DESCRIPTION)
@command_arg('name')
@command_opt('until-phase', help='Runs until including the given phase but then exits')
@command_opt('until-actor', help='Runs until including the given actor but then exits')
@command_opt('save-output', is_flag=True,
             help='Saves the output for actors to be consumable when executed with snactor run')
@command_opt('--whitelist-experimental', action='append', metavar='ActorName',
             help='Enables experimental actors')
@requires_repository
def cli(params):
    def impl(context=None):
        start = datetime.datetime.utcnow()
        configure_logger()
        repository = find_and_scan_repositories(find_repository_basedir('.'), include_locals=True)
        try:
            repository.load()
        except LeappError as exc:
            sys.stderr.write(exc.message)
            sys.stderr.write('\n')
            sys.exit(1)

        wf = repository.lookup_workflow(params.name)
        if not wf:
            raise CommandError('Could not find any workflow named "{}"'.format(params.name))

        instance = wf()
        for actor_name in params.whitelist_experimental or ():
            actor = repository.lookup_actor(actor_name)
            if actor:
                instance.whitelist_experimental_actor(actor)

        with beautify_actor_exception():
            instance.run(context=context, until_phase=params.until_phase, until_actor=params.until_actor)

        report_errors(instance.errors)
        report_deprecations(os.getenv('LEAPP_EXECUTION_ID'), start=start)

        if instance.failure:
            sys.exit(1)

    @with_snactor_context
    def snactor_context_impl():
        impl(context=os.getenv('LEAPP_EXECUTION_ID'))

    if params.save_output:
        snactor_context_impl()
    else:
        impl()



from collections import namedtuple
import datetime
import json

import pytest

import leapp.utils.report

_CONTEXT = 'some_str'
curr_date = datetime.datetime.today()


def _create_entry(data, event='deprecation', context=_CONTEXT):
    return {
        'id': '1',
        'stamp': '2020-09-11T15:56:23.767457Z',
        'actor': 'actor',
        'hostname': 'hostname',
        'data': json.dumps(data),
        'context': context,
        'event': event,
    }


_AUDIT_DATA = (
    # first two entries should be filtered out
    _create_entry(context='ignore123', data={}),
    _create_entry(event='something', data={}),
    _create_entry({
        'message': 'Usage of deprecated Model "DeprecatedModel"',
        'filename': '/etc/leapp/repos.d/system_upgrade/el7toel8/actors/fooactor/actor.py',
        'line': '        self.produce(DeprecatedModel(foo=bar))\n',
        'lineno': 51,
        'since': curr_date.strftime('%Y-%m-%d'),
        'reason': 'The DeprecatedModel has been deprecated.',
    }),
    # this is one is duplicate of the previous one
    _create_entry({
        'message': 'Usage of deprecated Model "DeprecatedModel"',
        'filename': '/etc/leapp/repos.d/system_upgrade/el7toel8/actors/fooactor/actor.py',
        'line': '        self.produce(DeprecatedModel(foo=bar))\n',
        'lineno': 51,
        'since': curr_date.strftime('%Y-%m-%d'),
        'reason': 'The DeprecatedModel has been deprecated.',
    }),
    _create_entry({
        'message': 'Usage of deprecated Model "DetonatedModel"',
        'filename': '/etc/leapp/repos.d/system_upgrade/el7toel8/actors/fooactor/actor.py',
        'line': '        self.produce(DetonatedModel(foo=bar))\n',
        'lineno': 51,
        'since': curr_date.strftime('%Y-%m-%d'),
        'reason': 'The DetonatedModel has been deprecated.',
    }),
    # to have at least one old entry
    _create_entry({
        'message': 'Usage of deprecated Model "AModel"',
        'filename': '/etc/leapp/repos.d/system_upgrade/el7toel8/actors/fooactor/actor.py',
        'line': '        self.produce(Amodel(foo=bar))\n',
        'lineno': 51,
        'since': '2018-01-01',
        'reason': 'The Amodel has been deprecated.',
    }),
)


def mocked_get_audit_entry(event, context):
    return (entry for entry in _AUDIT_DATA if entry['context'] == context and entry['event'] == event)


def test_create_report_from_deprecations(monkeypatch):
    monkeypatch.setattr(leapp.utils.report, 'get_audit_entry', mocked_get_audit_entry)

    reports = leapp.utils.report._create_reports_from_deprecations(_CONTEXT)
    # no duplicates and only deprecated messages for the given context
    assert len(reports) == len(_AUDIT_DATA) - 3

    for report in reports:
        assert 'Usage of deprecated Model' in report['title']
        assert report['audience'] == 'developer'
        assert report['severity'] in ('high', 'medium')
    assert len([i for i in reports if i['severity'] == 'high']) == 1

    reports = leapp.utils.report._create_reports_from_deprecations('non-existing-context')
    assert not reports



from leapp.actors import Actor
from leapp.models import UnitTestConfig
from leapp.tags import UnitTestWorkflowTag


class ConfigProvider(Actor):
    """
    No documentation has been provided for the config_provider actor.
    """

    name = 'config_provider'
    consumes = ()
    produces = (UnitTestConfig,)
    tags = (UnitTestWorkflowTag,)

    def process(self):
        self.produce(UnitTestConfig())



import os

from leapp.messaging import BaseMessaging
from leapp.utils.audit import Message, Audit, MessageData, get_messages


class InProcessMessaging(BaseMessaging):
    """
    This class implements the direct database access for the messaging.
    """

    def _process_message(self, message):
        message['event'] = 'new-message'
        message_keys = ('stamp', 'topic', 'actor', 'phase', 'hostname', 'context', 'msg_type')
        audit_keys = ('event', 'stamp', 'data', 'actor', 'phase', 'hostname', 'context')
        message['msg_type'] = message.pop('type')
        payload = message.pop('message')
        msg = Message(**dict(((k, message[k]) for k in message_keys if k in message)))
        audit = Audit(**dict(((k, message[k]) for k in audit_keys if k in message)))
        audit.message = msg
        audit.message.data = MessageData(data=payload['data'], hash_id=payload['hash'])
        audit.store()
        return message

    def _perform_load(self, consumes):
        context = os.environ.get('LEAPP_EXECUTION_ID', 'TESTING-CONTEXT')
        self._data = get_messages([consume.__name__ for consume in consumes], context)



class Flags(object):
    restart_after_phase = False
    request_restart_after_phase = False
    is_checkpoint = False

    def __init__(self,
                 request_restart_after_phase=False,
                 restart_after_phase=False,
                 is_checkpoint=False):
        self.request_restart_after_phase = request_restart_after_phase
        self.restart_after_phase = restart_after_phase
        self.is_checkpoint = is_checkpoint

    def serialize(self):
        """
        :return: Serialized data of phase flags
        """
        return {
            'restart_after_phase': self.restart_after_phase,
            'request_restart_after_phase': self.request_restart_after_phase,
            'is_checkpoint': self.is_checkpoint
        }



import os

from leapp.repository import Repository, DefinitionKind
from leapp.repository.manager import RepositoryManager
from leapp.repository.actor_definition import ActorDefinition
from leapp.exceptions import RepositoryConfigurationError
from leapp.utils.repository import get_global_repositories_data, get_user_config_repo_data, find_repos


def _make_repo_lookup(include_locals):
    data = {}
    for repo_id, entry in get_global_repositories_data().items():
        if entry['enabled']:
            data.update({entry['id']: entry['path']})

    if include_locals:
        # Having it here allows to override global repositories with local ones.
        data.update(get_user_config_repo_data().get('repos', {}))

    return data


def _resolve_repository_links(manager, include_locals):
    repo_lookup = _make_repo_lookup(include_locals=include_locals)
    if not repo_lookup and manager.get_missing_repo_links():
        # No repositories configured at all though missing links present
        raise RepositoryConfigurationError('No repos configured? Try adding some with "snactor repo find"')
    finished = False
    while not finished:
        missing = manager.get_missing_repo_links()
        for repo_id in missing:
            if repo_id in repo_lookup:
                manager.add_repo(scan_repo(repo_lookup[repo_id]))
                break
        else:
            finished = True

    if manager.get_missing_repo_links():
        raise RepositoryConfigurationError('Missing repositories detected: {}'.format(', '.join(missing)))


def find_and_scan_repositories(path, manager=None, include_locals=False):
    """
    Finds and scans all repositories found in the path and it will also resolve linked repositories.
    Using include_locals=True will additionally include user local repositories to be considered for
    resolving linked repositories.

    :param path: Path to scan for repositories
    :param manager: Optional repository manager to add found repos too
    :param include_locals: Should repositories linked be searched from the user local registry
    :return: repository manager instance (either passed through or a new instance if none was passed)
    """
    if os.path.exists(path):
        manager = manager or RepositoryManager()
        for repository in find_repos(path):
            manager.add_repo(scan_repo(repository))
        _resolve_repository_links(manager=manager, include_locals=include_locals)
    return manager


def scan_repo(path):
    """
    Scans all related repository resources

    :param path:
    :type path: str
    :return: repository
    """
    path = os.path.abspath(path)
    return scan(Repository(path), path)


def scan(repository, path):
    """
    Scans all repository resources

    :param repository:
    :type repository: :py:class:`leapp.repository.Repository`
    :param path: path to the repository
    :type path: str
    :return: instance of :py:class:`leapp.repository.Repository`
    """
    repository.log.debug("Scanning path %s", path)
    scan_tasks = (
        ('topics', scan_topics),
        ('models', scan_models),
        ('actors', scan_actors),
        ('tags', scan_tags),
        ('workflows', scan_workflows),
        ('files', scan_files),
        ('libraries', scan_libraries),
        ('tests', scan_tests),
        ('tools', scan_tools),
        ('apis', scan_apis))

    dirs = [e for e in os.listdir(path) if os.path.isdir(os.path.join(path, e))]
    for name, task in scan_tasks:
        if name in dirs:
            task(repository, os.path.join(path, name), path)
    return repository


def scan_topics(repo, path, repo_path):
    """
    Scans topics and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the topics
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    for root, unused, files in os.walk(path):
        for module in files:
            unused, ext = os.path.splitext(module)
            if ext == '.py':
                path = os.path.join(root, module)
                repo.add(DefinitionKind.TOPIC, os.path.relpath(path, repo_path))


def scan_actors(repo, path, repo_path):
    """
    Scans actors and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the actors
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    for root, unused, files in os.walk(path):
        for module in files:
            if module == 'actor.py':
                rel_path = os.path.relpath(root, repo_path)
                repo.add(DefinitionKind.ACTOR, scan(ActorDefinition(rel_path, repo_path, log=repo.log), root))


def scan_tags(repo, path, repo_path):
    """
    Scans tags and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the tags
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    for root, unused, files in os.walk(path):
        for module in files:
            unused, ext = os.path.splitext(module)
            if ext == '.py':
                path = os.path.join(root, module)
                repo.add(DefinitionKind.TAG, os.path.relpath(path, repo_path))


def scan_models(repo, path, repo_path):
    """
    Scans models and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the models
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    for root, unused, files in os.walk(path):
        for module in files:
            unused, ext = os.path.splitext(module)
            if ext == '.py':
                path = os.path.join(root, module)
                repo.add(DefinitionKind.MODEL, os.path.relpath(path, repo_path))


def scan_workflows(repo, path, repo_path):
    """
    Scans workflows and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the workflows
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    for root, unused, files in os.walk(path):
        for module in files:
            unused, ext = os.path.splitext(module)
            if ext == '.py':
                path = os.path.join(root, module)
                repo.add(DefinitionKind.WORKFLOW, os.path.relpath(path, repo_path))


def scan_files(repo, path, repo_path):
    """
    Scans files and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the files
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    if os.listdir(path):
        repo.add(DefinitionKind.FILES, os.path.relpath(path, repo_path))


def scan_libraries(repo, path, repo_path):
    """
    Scans libraries and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the libraries
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    if os.listdir(path):
        repo.add(DefinitionKind.LIBRARIES, os.path.relpath(path, repo_path))


def scan_tools(repo, path, repo_path):
    """
    Scans tools and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the tools
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    if os.listdir(path):
        repo.add(DefinitionKind.TOOLS, os.path.relpath(path, repo_path))


def scan_tests(repo, path, repo_path):
    """
    Scans tests and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the tests
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    if os.listdir(path):
        repo.add(DefinitionKind.TESTS, os.path.relpath(path, repo_path))


def scan_apis(repo, path, repo_path):
    """
    Scans apis and adds them to the repository.

    :param repo: Instance of the repository
    :type repo: :py:class:`leapp.repository.Repository`
    :param path: path to the apis
    :type path: str
    :param repo_path: path to the repository
    :type repo_path: str
    """
    if os.listdir(path):
        repo.add(DefinitionKind.API, os.path.relpath(path, repo_path))



from leapp.actors import Actor
from leapp.models import DepCheck3, DepCheck1
from leapp.tags import FirstPhaseTag, WorkflowApiTestWorkflowTag


class DepCheckProduces(Actor):
    """
    Produces messages DepCheck1 and DepCheck3 which are going to be consumed by the DepCheckAPI1 and DepCheckAPI3 APIs.
    """

    name = 'dep_check_produces'
    consumes = ()
    produces = (DepCheck1, DepCheck3)
    tags = (FirstPhaseTag, WorkflowApiTestWorkflowTag)

    def process(self):
        self.produce(DepCheck1(), DepCheck3())



import getpass
import json
import os
import sqlite3
import uuid

from leapp.utils.audit import get_connection, Execution, Host, MessageData, \
    DataSource, Message, Audit, get_messages, checkpoint, get_checkpoints, create_audit_entry, get_audit_entry
from leapp.config import get_config
from leapp.libraries.stdlib import run

_HOSTNAME = 'test-host.example.com'
_CONTEXT_NAME = 'test-context-name'
_ACTOR_NAME = 'test-actor-name'
_PHASE_NAME = 'test-phase-name'
_MESSAGE_TYPE = 'MessageType'
_TOPIC_NAME = 'test-topic'

_ORIGINAL_DB_SCHEMA = '''
CREATE TABLE execution (
  id            INTEGER PRIMARY KEY NOT NULL,
  context       VARCHAR(36)         NOT NULL UNIQUE,
  stamp         TIMESTAMP           NOT NULL DEFAULT CURRENT_TIMESTAMP,
  configuration TEXT                         DEFAULT NULL
);

CREATE TABLE host (
  id       INTEGER PRIMARY KEY NOT NULL,
  context  VARCHAR(36)         NOT NULL REFERENCES execution (context),
  hostname VARCHAR(255)        NOT NULL,
  UNIQUE (context, hostname)
);

CREATE TABLE message_data (
  hash VARCHAR(64) PRIMARY KEY NOT NULL,
  data TEXT
);

CREATE TABLE data_source (
  id      INTEGER PRIMARY KEY NOT NULL,
  context VARCHAR(36)         NOT NULL REFERENCES execution (context),
  host_id INTEGER             NOT NULL REFERENCES host (id),
  actor   VARCHAR(1024)       NOT NULL DEFAULT '',
  phase   VARCHAR(1024)       NOT NULL DEFAULT '',
  UNIQUE (context, host_id, actor, phase)
);


CREATE TABLE message (
  id                INTEGER PRIMARY KEY NOT NULL,
  context           VARCHAR(36)         NOT NULL REFERENCES execution (context),
  stamp             TIMESTAMP           NOT NULL DEFAULT CURRENT_TIMESTAMP,
  channel           VARCHAR(1024)       NOT NULL,
  type              VARCHAR(1024)       NOT NULL,
  data_source_id    INTEGER             NOT NULL REFERENCES data_source (id),
  message_data_hash VARCHAR(64)         NOT NULL REFERENCES message_data (hash)
);


CREATE TABLE audit (
  id             INTEGER PRIMARY KEY NOT NULL,
  event          VARCHAR(256)        NOT NULL REFERENCES execution (context),
  stamp          TIMESTAMP           NOT NULL DEFAULT CURRENT_TIMESTAMP,
  context        VARCHAR(36)         NOT NULL,
  data_source_id INTEGER             NOT NULL REFERENCES data_source (id),

  message_id     INTEGER                      DEFAULT NULL REFERENCES message (id),
  data           TEXT                         DEFAULT NULL
);


CREATE VIEW messages_data AS
  SELECT
    message.id        AS id,
    message.context   AS context,
    message.stamp     AS stamp,
    message.channel   AS channel,
    message.type      as type,
    data_source.actor as actor,
    data_source.phase as phase,
    msg_data.hash     as message_hash,
    msg_data.data     as message_data,
    host.hostname     as hostname
  FROM
    message
    JOIN
    data_source ON data_source.id = message.data_source_id
    ,
    message_data as msg_data ON message.message_data_hash = msg_data.hash,
  host ON host.id = data_source.host_id;
'''


def setup_module():
    get_config().set('database', 'path', '/tmp/leapp-test.db')


def setup():
    path = get_config().get('database', 'path')
    if os.path.isfile(path):
        os.unlink(path)


def test_migrations_are_applied():
    con = sqlite3.connect(get_config().get('database', 'path'))
    con.executescript(_ORIGINAL_DB_SCHEMA)
    assert con.execute("PRAGMA user_version").fetchone()[0] == 0
    con.close()

    with get_connection(None) as db:
        assert db.execute("PRAGMA user_version").fetchone()[0] > 0


def test_pass_through():
    with get_connection(None) as db:
        with get_connection(db) as db2:
            assert db is db2


def test_execution():
    e = Execution(context=_CONTEXT_NAME, configuration=json.dumps({'data': 'nothing to store'}))
    e.store()
    assert e.execution_id


def test_host():
    e = Host(context=_CONTEXT_NAME, hostname=_HOSTNAME)
    e.store()
    assert e.host_id


def test_message_data(saved=True):
    e = MessageData(data='abc', hash_id='abc')
    if saved:
        e.store()
        with get_connection(None) as conn:
            data = conn.execute(
                'SELECT data from message_data WHERE hash = "abc"').fetchone()[0]
        assert data == 'abc'
    return e


def test_data_source():
    e = DataSource(actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME, hostname=_HOSTNAME)
    e.store()
    assert e.data_source_id
    assert e.host_id
    return e


def test_message(saved=True):
    data = test_message_data(saved=True)
    e = Message(actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME, hostname=_HOSTNAME,
                topic=_TOPIC_NAME, msg_type=_MESSAGE_TYPE, data=data)
    if saved:
        e.store()
        assert e.message_id
        assert e.host_id
        assert e.data_source_id
    return e


def test_message_not_saved_data():
    data = test_message_data(saved=False)
    e = Message(actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME, hostname=_HOSTNAME,
                topic=_TOPIC_NAME, msg_type=_MESSAGE_TYPE, data=data)
    e.store()
    assert e.message_id
    assert e.host_id
    assert e.data_source_id


def test_audit_message():
    msg = test_message(saved=True)
    e = Audit(event='new-message', message=msg, actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME,
              hostname=_HOSTNAME)
    e.store()
    assert e.data_source_id
    assert e.host_id
    assert e.message.message_id
    assert e.message.data_source_id == e.data_source_id
    assert e.message.host_id == e.host_id


def test_audit_not_saved_message():
    msg = test_message(saved=False)
    e = Audit(event='new-message', message=msg, actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME,
              hostname=_HOSTNAME)
    e.store()
    assert e.data_source_id
    assert e.host_id
    assert e.message.message_id
    assert e.message.data_source_id == e.data_source_id
    assert e.message.host_id == e.host_id


def test_audit_data():
    e = Audit(event='new-message', data='Some data', actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME,
              hostname=_HOSTNAME)
    e.store()
    assert e.data_source_id
    assert e.host_id
    assert not e.message


def test_audit_non_string_data():
    e = Audit(event='new-message', data=['Some data'], actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME,
              hostname=_HOSTNAME)
    e.store()
    assert e.audit_id
    current = e.audit_id
    e.store()
    assert current is e.audit_id
    assert e.data_source_id
    assert e.host_id
    assert not e.message


def test_get_messages():
    assert not get_messages((), _CONTEXT_NAME)
    messages = get_messages((_MESSAGE_TYPE,), _CONTEXT_NAME)
    assert messages is not None and not messages

    test_message()
    messages = get_messages((_MESSAGE_TYPE,), _CONTEXT_NAME)
    assert messages and len(messages) == 1


def test_checkpoints():
    checkpoint(actor=_ACTOR_NAME, phase=_PHASE_NAME, context=_CONTEXT_NAME, hostname=_HOSTNAME)
    result = get_checkpoints(_CONTEXT_NAME)
    assert result and len(result) == 1
    assert result[0]['id']
    assert result[0]['actor'] == _ACTOR_NAME
    assert result[0]['phase'] == _PHASE_NAME
    assert result[0]['stamp'].endswith('Z')


def test_create_audit_entry(monkeypatch):
    monkeypatch.setenv('LEAPP_CURRENT_ACTOR', _ACTOR_NAME)
    monkeypatch.setenv('LEAPP_CURRENT_PHASE', _PHASE_NAME)
    monkeypatch.setenv('LEAPP_EXECUTION_ID', _CONTEXT_NAME)
    monkeypatch.setenv('LEAPP_HOSTNAME', _HOSTNAME)
    _id = str(uuid.uuid4())
    event = 'process-start'
    create_audit_entry(event, {'id': _id, 'parameters': 'ls'})
    assert get_audit_entry(event, _CONTEXT_NAME)


def test_audit_command_in_db(monkeypatch):
    monkeypatch.setenv('LEAPP_CURRENT_ACTOR', _ACTOR_NAME)
    monkeypatch.setenv('LEAPP_CURRENT_PHASE', _PHASE_NAME)
    monkeypatch.setenv('LEAPP_EXECUTION_ID', _CONTEXT_NAME)
    monkeypatch.setenv('LEAPP_HOSTNAME', _HOSTNAME)
    _id = str(uuid.uuid4())
    cmd = ['whoami']
    monkeypatch.setattr(uuid, 'uuid4', lambda: _id)
    result = run(cmd)
    assert result['stdout'] == '{}\n'.format(getpass.getuser())
    event = 'process-start'
    assert get_audit_entry(event, _CONTEXT_NAME)
    event = 'process-result'
    assert get_audit_entry(event, _CONTEXT_NAME)



from leapp.workflows import Workflow
from leapp.workflows.phases import Phase
from leapp.workflows.flags import Flags
from leapp.workflows.tagfilters import TagFilter
from leapp.workflows.policies import Policies
from leapp.tags import FirstPhaseTag, SecondPhaseTag, WorkflowApiTestWorkflowTag


class WorkflowApiTestWorkflow(Workflow):
    name = 'WorkflowAPITest'
    tag = WorkflowApiTestWorkflowTag
    short_name = 'workflow_api_test'
    description = '''No description has been provided for the WorkflowAPITest workflow.'''

    class FirstPhase(Phase):
        name = 'first_phase'
        filter = TagFilter(FirstPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()

    class SecondPhase(Phase):
        name = 'second_phase'
        filter = TagFilter(SecondPhaseTag)
        policies = Policies(Policies.Errors.FailPhase,
                            Policies.Retry.Phase)
        flags = Flags()



class Policies(object):
    class Errors(object):
        class FailPhase(object):
            pass

        class FailImmediately(object):
            pass

        class ReportOnly(object):
            pass

    class Retry(object):
        class Actor(object):
            pass

        class Phase(object):
            pass

        class Disabled(object):
            pass

    def __init__(self, error=Errors.ReportOnly, retry=Retry.Disabled):
        self.error = error
        self.retry = retry

    def serialize(self):
        return {
            'error': self.error.__name__,
            'retry': self.retry.__name__
        }



import os
import uuid

from leapp.utils.audit import get_connection, Execution
from leapp.utils.clicmd import command_aware_wraps


def last_snactor_context(connection=None):
    """
    Retrieves the last snactor-run context from the database. It generates a new one if none has been found.

    :param connection: Database connection to use instead of the default connection.
    :returns: String representing the latest snactor-run context uuid.
    """
    with get_connection(db=connection) as db:
        cursor = db.execute('''
            SELECT context, stamp FROM execution WHERE kind = 'snactor-run' ORDER BY id DESC LIMIT 1
        ''')
        row = cursor.fetchone()
        if row:
            context = row[0]
        else:
            context = str(uuid.uuid4())
            Execution(context=context, kind='snactor-run', configuration='').store()
        return context


def with_snactor_context(f):
    @command_aware_wraps(f)
    def wrapper(*args, **kwargs):
        os.environ["LEAPP_EXECUTION_ID"] = last_snactor_context()
        return f(*args, **kwargs)
    return wrapper



import itertools

from leapp.models import resolve_model_references
from leapp.repository import _LoadStage


class RepositoryManager(object):
    """
    Handles multiple loaded repositories
    """

    def __init__(self):
        self._repos = {}

    def lookup_actor(self, name):
        """
        Find actor in all loaded repositories

        :param name: Name of the actor
        :type name: str
        :return: None or Actor
        """
        for repo in self._repos.values():
            actor = repo.lookup_actor(name)
            if actor:
                return actor
        return None

    def lookup_workflow(self, name):
        """
        Find workflow in all loaded repositories

        :param name: Name of the workflow
        :type name: str
        :return: None or Workflow
        """
        for repo in self._repos.values():
            workflow = repo.lookup_workflow(name)
            if workflow:
                return workflow
        return None

    def get_missing_repo_links(self):
        """
        Gather all missing repository ids linked by the added repositories.

        :return: Set of missing repository ids.
        """
        missing = set()
        available = set(self._repos.keys())
        for repo in self._repos.values():
            missing.update(set(repo.repo_links).difference(available))
        return missing

    def add_repo(self, repo):
        """
        Add new repository to manager.

        :param repo: Repository to be added (registered)
        :type repo: :py:class:`leapp.repository.Repository`
        """
        self._repos[repo.repo_id] = repo

    @property
    def repos(self):
        """
        :return: A tuple of all repository instances
        """
        return tuple(self._repos.values())

    def repo_by_id(self, repo_id):
        """
        Look up a repository by id

        :param repo_id: Repository id
        :return: Repository or None
        """
        return self._repos.get(repo_id, None)

    def load(self, resolve=True, skip_actors_discovery=False):
        """
        Load all known repositories.

        :param resolve: Whether or not to perform the resolving of model references
        :type resolve: bool
        :param skip_actors_discovery: specifies whether to skip discovery process of the actors
            When we testing actors, we're directly injecting the actor context, so we don't
            need to inject it during the repo loading. This option helps to solve this problem.
        :type skip_actors_discovery: bool
        """
        for repo in self._repos.values():
            repo.load(resolve=False, stage=_LoadStage.INITIAL)

        for repo in self._repos.values():
            repo.load(resolve=False, stage=_LoadStage.MODELS)

        for repo in self._repos.values():
            repo.load(resolve=False, stage=_LoadStage.LIBRARIES)

        for repo in self._repos.values():
            repo.load(resolve=False, stage=_LoadStage.ACTORS, skip_actors_discovery=skip_actors_discovery)

        for repo in self._repos.values():
            repo.load(resolve=False, stage=_LoadStage.WORKFLOWS)

        if resolve:
            resolve_model_references()

    def serialize(self):
        """
        :return: List of resources in all known repositories
        """
        return [repo.serialize() for repo in self._repos.values()]

    @property
    def actors(self):
        """
        :return: Tuple of :py:class:`leapp.repository.actor_definition.ActorDefinition` instances representing actors \
        from all repositories
        """
        return tuple(itertools.chain(*[repo.actors for repo in self._repos.values()]))

    @property
    def topics(self):
        """
        :return: Tuple of paths to topic-defining python modules from all repositories
        """
        return tuple(itertools.chain(*[repo.topics for repo in self._repos.values()]))

    @property
    def models(self):
        """
        :return: Tuple of paths to model-defining python modules from all repositories
        """
        return tuple(itertools.chain(*[repo.models for repo in self._repos.values()]))

    @property
    def tags(self):
        """
        :return: Tuple of paths to tag-defining python modules from all repositories
        """
        return tuple(itertools.chain(*[repo.tags for repo in self._repos.values()]))

    @property
    def workflows(self):
        """
        :return: Tuple of paths to workflow-defining python modules from all repositories
        """
        return tuple(itertools.chain(*[repo.workflows for repo in self._repos.values()]))

    @property
    def tools(self):
        """
        :return: Tuple of paths to "tools" folders from all repositories
        """
        return tuple(itertools.chain(*[repo.tools for repo in self._repos.values()]))

    @property
    def libraries(self):
        """
        :return: Tuple of paths to "libraries" folders from all repositories
        """
        return tuple(itertools.chain(*[repo.libraries for repo in self._repos.values()]))

    @property
    def files(self):
        """
        :return: Tuple of paths to "files" folders from all repositories
        """
        return tuple(itertools.chain(*[repo.files for repo in self._repos.values()]))



import pytest

import leapp.messaging.answerstore
from leapp.messaging.answerstore import AnswerStore, multiprocessing


class MockManager(object):
    def __call__(self):
        return self

    @staticmethod
    def dict(*args):
        return dict(*args)


class MockComponentKey1(object):
    value_type = str
    key = 'key1'
    label = 'label key1'
    description = 'description key1'
    default = False
    value = None


class MockComponentKey2(object):
    value_type = str
    key = 'key2'
    label = 'label key2'
    description = 'description key2'
    default = 'Default'
    value = None


class MockComponentKey1Bool(object):
    value_type = bool
    key = 'key1'
    label = 'label bool key1'
    description = 'description bool key1'
    default = True
    value = None


class MockComponentKey2Bool(object):
    value_type = bool
    key = 'key2'
    label = 'label bool key2'
    description = 'description bool key2'
    default = False
    value = None


class MockDialogScope1(object):
    scope = 'scope1'
    title = 'Mock Dialog scope1'
    reason = 'Mock Test Data for scope1'
    components = (MockComponentKey1, MockComponentKey2)


class MockDialogScope2(object):
    scope = 'scope2'
    title = 'Mock Dialog scope2'
    reason = 'Mock Test Data for scope2'
    components = (MockComponentKey1, MockComponentKey2)


class MockDialogBoolScope(object):
    scope = 'boolscope'
    title = 'Mock Dialog boolscope'
    reason = 'Mock Test Data for boolscope'
    components = (MockComponentKey1Bool, MockComponentKey2Bool)


class MockWorkflow(object):
    dialogs = (MockDialogScope1, MockDialogScope2, MockDialogBoolScope)


@pytest.fixture(scope='function')
def answerfile_data():
    yield '''
[scope1]
key1=scope1.key1
key2=scope1.key2

[scope2]
key1=scope2.key1
key2=scope2.key2

[boolscope]
key1=True
Key2=False
    '''


@pytest.fixture(scope='function', name='answerfile')
def answerfile_file(tmpdir, answerfile_data):
    d = tmpdir.join('answerfile')
    d.mkdir()
    f = d.join('data.ini')
    f.write(answerfile_data)
    yield str(f)


def test_answerstore_init(monkeypatch):
    manager = MockManager()
    assert type(AnswerStore()._storage) is multiprocessing.managers.DictProxy
    monkeypatch.setattr(multiprocessing, 'Manager', MockManager())
    assert type(AnswerStore(manager=manager)._storage) is dict
    assert type(AnswerStore()._storage) is dict


def test_answerstore_answer():
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.answer(scope='scope1', key='key1', value='scope1.key1')
    a.answer(scope='scope1', key='key2', value='scope1.key2')
    a.answer(scope='scope2', key='key1', value='scope2.key1')
    a.answer(scope='scope2', key='key2', value='scope2.key2')
    assert 'scope1' in store.keys()
    assert 'scope2' in store.keys()
    assert 'key1' in store['scope1'].keys()
    assert 'key1' in store['scope1'].keys()
    assert 'key1' in store['scope2']
    assert 'key2' in store['scope2']
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'


def test_answerstore_update(answerfile, answerfile_data, tmpdir):
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load(answerfile)
    assert store['boolscope']['key1'] == 'True'
    assert store['boolscope']['key2'] == 'False'
    a.answer('boolscope', 'key1', False)
    a.answer('boolscope', 'key2', True)
    assert store['boolscope']['key1'] is False
    assert store['boolscope']['key2'] is True
    f = tmpdir.join('test_answerstore_update')
    f.write(answerfile_data)
    a.update(str(f))
    a2 = AnswerStore(manager=m)
    store2 = a2._storage
    a2.load(str(f))
    assert store2['boolscope']['key1'] == 'False'
    assert store2['boolscope']['key2'] == 'True'
    assert store2['scope1']['key1'] == 'scope1.key1'
    assert store2['scope1']['key2'] == 'scope1.key2'
    assert store2['scope2']['key1'] == 'scope2.key1'
    assert store2['scope2']['key2'] == 'scope2.key2'


def test_answerstore_load(answerfile):
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load(answerfile)
    assert 'scope1' in store.keys()
    assert 'scope2' in store.keys()
    assert 'boolscope' in store.keys()
    assert 'key1' in store['scope1'].keys()
    assert 'key1' in store['scope1'].keys()
    assert 'key1' in store['scope2']
    assert 'key2' in store['scope2']
    assert 'key1' in store['boolscope']
    assert 'key2' in store['boolscope']
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'
    assert store['boolscope']['key1'] == 'True'
    assert store['boolscope']['key2'] == 'False'


def test_answerstore_load_and_translate_for_workflow(answerfile):
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load_and_translate_for_workflow(answerfile, MockWorkflow)
    assert store['boolscope']['key1'] is True
    assert store['boolscope']['key2'] is False
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'


def test_answerfile_get(monkeypatch, answerfile):
    entries_created = []

    def mocked_create_audit_entry(event, data, message=None):
        entries_created.append((event, data, message))

    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load(answerfile)
    monkeypatch.setattr(leapp.messaging.answerstore, 'create_audit_entry', mocked_create_audit_entry)

    # Testing boolscope
    assert store['boolscope']['key1'] == 'True'
    assert store['boolscope']['key2'] == 'False'
    result = a.get(scope='boolscope', fallback='boolscope')
    assert result != 'boolscope'
    assert result['key1'] == 'True'
    assert result['key2'] == 'False'
    assert entries_created
    assert entries_created[0][1]['scope'] == 'boolscope'
    assert entries_created[0][1]['fallback'] == 'boolscope'
    assert entries_created[0][1]['answer']['key1'] == 'True'
    assert entries_created[0][1]['answer']['key2'] == 'False'
    entries_created.pop()

    # Testing scope1
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    result = a.get(scope='scope1', fallback='scope1')
    assert result
    assert result['key1'] == 'scope1.key1'
    assert result['key2'] == 'scope1.key2'
    assert entries_created
    assert entries_created[0][1]['scope'] == 'scope1'
    assert entries_created[0][1]['fallback'] == 'scope1'
    assert entries_created[0][1]['answer']['key1'] == 'scope1.key1'
    assert entries_created[0][1]['answer']['key2'] == 'scope1.key2'
    entries_created.pop()

    # Testing scope2
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'

    result = a.get(scope='scope2', fallback='scope2')
    assert result
    assert result['key1'] == 'scope2.key1'
    assert result['key2'] == 'scope2.key2'
    assert entries_created
    assert entries_created[0][1]['scope'] == 'scope2'
    assert entries_created[0][1]['fallback'] == 'scope2'
    assert entries_created[0][1]['answer']['key1'] == 'scope2.key1'
    assert entries_created[0][1]['answer']['key2'] == 'scope2.key2'
    entries_created.pop()


def test_answerfile_translate_for_workflow(answerfile):
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load(answerfile)
    assert store['boolscope']['key1'] == 'True'
    assert store['boolscope']['key2'] == 'False'
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'
    a.translate_for_workflow(MockWorkflow)
    assert store['boolscope']['key1'] is True
    assert store['boolscope']['key2'] is False
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'


def test_answerfile_translate(answerfile):
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load(answerfile)
    assert store['boolscope']['key1'] == 'True'
    assert store['boolscope']['key2'] == 'False'
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'
    a.translate(MockDialogBoolScope)
    assert store['boolscope']['key1'] is True
    assert store['boolscope']['key2'] is False
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'


def test_answerfile_generate(tmpdir, answerfile):
    f = tmpdir.join('test_answerfile_generate')
    m = MockManager()
    a = AnswerStore(manager=m)
    store = a._storage
    a.load_and_translate_for_workflow(answerfile, MockWorkflow)
    assert store['scope1']['key1'] == 'scope1.key1'
    assert store['scope1']['key2'] == 'scope1.key2'
    assert store['scope2']['key1'] == 'scope2.key1'
    assert store['scope2']['key2'] == 'scope2.key2'
    assert store['boolscope']['key1'] is True
    assert store['boolscope']['key2'] is False
    a.answer('scope1', 'key1', 'generate.scope1.key1')
    a.answer('scope1', 'key2', 'generate.scope1.key2')
    a.answer('scope2', 'key1', 'generate.scope2.key1')
    a.answer('scope2', 'key2', 'generate.scope2.key2')
    a.answer('boolscope', 'key1', False)
    a.answer('boolscope', 'key2', True)
    a.generate(MockWorkflow.dialogs, str(f))
    a = AnswerStore(manager=m)
    store2 = a._storage
    a.load_and_translate_for_workflow(str(f), MockWorkflow)
    assert store2['scope1']['key1'] == 'generate.scope1.key1'
    assert store2['scope1']['key2'] == 'generate.scope1.key2'
    assert store2['scope2']['key1'] == 'generate.scope2.key1'
    assert store2['scope2']['key2'] == 'generate.scope2.key2'
    assert store2['boolscope']['key1'] is False
    assert store2['boolscope']['key2'] is True



import pkgutil


class LeappLibrariesFinder(object):
    """
    Implements functionality to dynamically load libraries for actors.
    """

    def __init__(self, module_prefix, paths):
        """
        :param module_prefix: Prefix string such as 'leapp.libraries.common' or 'leapp.libraries.actor' which is used
                              to filter the modules to handle with this finder.
        :type module_prefix: str
        :param paths: List of paths to search for the matching libraries
        :type paths: List or Tuple
        """
        self._paths = paths
        self._prefix = module_prefix

    def _implementation(self, method, fullname, path):  # noqa; pylint: disable=unused-argument
        if not fullname.startswith(self._prefix + '.'):
            return None
        module = fullname.split('.')[-1]
        for loader, name, ispkg in pkgutil.iter_modules(self._paths):
            if name == module:
                return getattr(loader, method)(fullname)

    def find_spec(self, fullname, path, target=None):  # noqa; pylint: disable=unused-argument
        """ Implementation for python >=3.4 """
        return self._implementation(method='find_spec', fullname=fullname, path=path)

    def find_module(self, fullname, path=None):
        """ Implementation for python <3.4 """
        return self._implementation(method='find_module', fullname=fullname, path=path)
